import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from ...tools import joint_normal, linear, power
from .. import CCA


class TestCCAStat:
    @pytest.mark.parametrize("n", [10, 100, 1000])
    @pytest.mark.parametrize("obs_stat", [1.0])
    @pytest.mark.parametrize("obs_pvalue", [1 / 1000])
    def test_linear_oned(self, n, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y = linear(n, 1)
        stat, pvalue = CCA().test(x, y)

        assert_almost_equal(stat, obs_stat, decimal=2)
        assert_almost_equal(pvalue, obs_pvalue, decimal=2)

    @pytest.mark.parametrize("n", [100, 1000, 10000])
    @pytest.mark.parametrize("obs_stat", [0.07])
    @pytest.mark.parametrize("obs_pvalue", [1 / 1000])
    def test_linear_threed(self, n, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y = joint_normal(n, 3)
        stat, pvalue = CCA().test(x, y)

        assert_almost_equal(stat, obs_stat, decimal=1)
        assert_almost_equal(pvalue, obs_pvalue, decimal=1)

    @pytest.mark.parametrize("n", [1000, 10000])
    def test_rep(self, n):
        x, y = joint_normal(n, 3)
        stat, pvalue = CCA().test(x, y, random_state=2)
        stat2, pvalue2 = CCA().test(x, y, random_state=2)

        assert stat == stat2
        assert pvalue == pvalue2


class TestCCATypeIError:
    def test_oned(self):
        np.random.seed(123456789)
        est_power = power(
            "CCA",
            sim_type="indep",
            sim="multimodal_independence",
            n=1000,
            p=1,
            alpha=0.05,
        )

        assert_almost_equal(est_power, 0.05, decimal=2)



import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_raises, assert_warns

from ...tools import linear, power
from .. import Dcorr


class TestDcorrStat:
    @pytest.mark.parametrize("n", [100, 200])
    @pytest.mark.parametrize("obs_stat", [1.0])
    @pytest.mark.parametrize("obs_pvalue", [1 / 1000])
    def test_linear_oned(self, n, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y = linear(n, 1)
        stat1, pvalue1 = Dcorr().test(x, y)
        stat2 = Dcorr().statistic(x, y)

        assert_almost_equal(stat1, obs_stat, decimal=2)
        assert_almost_equal(stat2, obs_stat, decimal=2)
        assert_almost_equal(pvalue1, obs_pvalue, decimal=2)

    @pytest.mark.parametrize("n", [100, 200])
    def test_rep(self, n):
        x, y = linear(n, 1)
        stat1, pvalue1 = Dcorr().test(x, y, random_state=2)
        stat2, pvalue2 = Dcorr().test(x, y, random_state=2)

        assert stat1 == stat2
        assert pvalue1 == pvalue2

    def test_dcorr_sqrt_bug(self):
        x = np.array([1,2,3,4,5])
        y = np.array([1,2,9,4,4])
        stat = Dcorr(bias=True).test(x, y, reps=0)[0]

        assert_almost_equal(stat, 0.762676242417, decimal=2)


class TestDcorrTypeIError:
    def test_oned(self):
        np.random.seed(123456789)
        est_power = power(
            "Dcorr",
            sim_type="indep",
            sim="multimodal_independence",
            n=100,
            p=1,
            alpha=0.05,
        )

        assert_almost_equal(est_power, 0.05, decimal=2)

    def test_oned_fast(self):
        np.random.seed(123456789)
        est_power = power(
            "Dcorr",
            sim_type="indep",
            sim="multimodal_independence",
            n=100,
            p=1,
            alpha=0.05,
            auto=True,
        )

        assert_almost_equal(est_power, 0.05, decimal=2)

    def test_threed_fast(self):
        np.random.seed(123456789)
        est_power = power(
            "Dcorr",
            sim_type="indep",
            sim="multimodal_independence",
            n=100,
            p=3,
            alpha=0.05,
            auto=True,
        )

        assert_almost_equal(est_power, 0.05, decimal=2)



import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from ...tools import indep_normal, power
from .. import PartialDcorr


class TestPDcorrStat:
    @pytest.mark.parametrize("n", [100, 200])
    @pytest.mark.parametrize("obs_stat", [0.0])
    def test_linear_oned(self, n, obs_stat):
        np.random.seed(123456789)
        x, y, z = indep_normal(n, 1)
        stat1, pvalue1 = PartialDcorr().test(x, y, z)
        stat2 = PartialDcorr().statistic(x, y, z)

        assert_almost_equal(stat1, obs_stat, decimal=2)
        assert_almost_equal(stat2, obs_stat, decimal=2)

    @pytest.mark.parametrize("n", [100, 200])
    def test_rep(self, n):
        x, y, z = indep_normal(n, 1)
        stat1, pvalue1 = PartialDcorr().test(x, y, z, random_state=2)
        stat2, pvalue2 = PartialDcorr().test(x, y, z, random_state=2)

        assert stat1 == stat2
        assert pvalue1 == pvalue2

    @pytest.mark.parametrize("n", [100])
    @pytest.mark.parametrize("obs_stat", [0.0143057])
    @pytest.mark.parametrize("obs_pvalue", [0.12])
    def test_indep_normal_corr(self, n, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y, z = indep_normal(n, 1)
        stat1, pvalue1 = PartialDcorr(use_cov=False).test(x, y, z)

        assert_almost_equal(stat1, obs_stat, decimal=2)
        assert_almost_equal(pvalue1, obs_pvalue, decimal=2)


class TestPDcorrTypeIError:
    def test_oned(self):
        np.random.seed(123456789)
        est_power = power(
            "PartialDcorr",
            sim_type="condi",
            sim="independent_normal",
            n=100,
            p=1,
            alpha=0.05,
        )
        assert_almost_equal(est_power, 0.05, decimal=2)



import numpy as np
from numba import jit
from scipy.sparse.construct import random

from ..tools import compute_dist
from ._utils import _CheckInputs
from .base import IndependenceTest, IndependenceTestOutput
from scipy.stats import rankdata


class HHG(IndependenceTest):
    r"""
    Heller Heller Gorfine (HHG) test statistic and p-value.

    This is a powerful test for independence based on calculating pairwise
    Euclidean distances and associations between these distance matrices. The
    test statistic is a function of ranks of these distances, and is
    consistent against similar tests
    :footcite:p:`hellerConsistentMultivariateTest2013`. It can also operate on multiple
    dimensions :footcite:p:`hellerConsistentMultivariateTest2013`.

    Parameters
    ----------
    compute_distance : str, callable, or None, default: "euclidean"
        A function that computes the distance among the samples within each
        data matrix.
        Valid strings for ``compute_distance`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,

            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.

        Set to ``None`` or ``"precomputed"`` if ``x`` and ``y`` are already distance
        matrices. To call a custom function, either create the distance matrix
        before-hand or create a function of the form ``metric(x, **kwargs)``
        where ``x`` is the data matrix for which pairwise distances are
        calculated and ``**kwargs`` are extra arguments to send to your custom
        function.
    **kwargs
        Arbitrary keyword arguments for ``compute_distance``.

    Notes
    -----
    The statistic can be derived as follows
    :footcite:p:`hellerConsistentMultivariateTest2013`:

    Let :math:`x` and :math:`y` be :math:`(n, p)` samples of random variables
    :math:`X` and :math:`Y`. For every sample :math:`j \neq i`, calculate the
    pairwise distances in :math:`x` and :math:`y` and denote this as
    :math:`d_x(x_i, x_j)` and :math:`d_y(y_i, y_j)`. The indicator function is
    denoted as :math:`\mathbb{1} \{ \cdot \}`. The cross-classification
    between these two random variables can be calculated as

    .. math::

        A_{11} = \sum_{k=1, k \neq i,j}^n
                    \mathbb{1} \{ d_x(x_i, x_k) \leq d_x(x_i, x_j) \}
                    \mathbb{1} \{ d_y(y_i, y_k) \leq d_y(y_i, y_j) \}

    and :math:`A_{12}`, :math:`A_{21}`, and :math:`A_{22}` are defined
    similarly. This is organized within the following table:

    +--------------------------------------------+--------------------------------------------+--------------------------------------------+---------------------------+
    |                                            | :math:`d_x(x_i, \cdot) \leq d_x(x_i, x_j)` | :math:`d_x(x_i, \cdot) \leq d_x(x_i, x_j)` |                           |
    +--------------------------------------------+--------------------------------------------+--------------------------------------------+---------------------------+
    | :math:`d_x(x_i, \cdot) \leq d_x(x_i, x_j)` | :math:`A_{11} (i,j)`                       | :math:`A_{12} (i,j)`                       | :math:`A_{1 \cdot} (i,j)` |
    +--------------------------------------------+--------------------------------------------+--------------------------------------------+---------------------------+
    | :math:`d_x(x_i, \cdot) > d_x(x_i, x_j)`    | :math:`A_{21} (i,j)`                       | :math:`A_{22} (i,j)`                       | :math:`A_{2 \cdot} (i,j)` |
    +--------------------------------------------+--------------------------------------------+--------------------------------------------+---------------------------+
    |                                            | :math:`A_{\cdot 1} (i,j)`                  | :math:`A_{\cdot 2} (i,j)`                  | :math:`n - 2`             |
    +--------------------------------------------+--------------------------------------------+--------------------------------------------+---------------------------+

    Here, :math:`A_{\cdot 1}` and :math:`A_{\cdot 2}` are the column sums,
    :math:`A_{1 \cdot}` and :math:`A_{2 \cdot}` are the row sums, and
    :math:`n - 2` is the number of degrees of freedom. From this table, we can
    calculate the Pearson's chi squared test statistic using,

    .. math::

        S(i, j) = \frac{(n-2) (A_{12} A_{21} - A_{11} A_{22})^2}
                       {A_{1 \cdot} A_{2 \cdot} A_{\cdot 1} A_{\cdot 2}}

    and the HHG test statistic is then,

    .. math::

        \mathrm{HHG}_n (x, y) = \sum_{i=1}^n \sum_{j=1, j \neq i}^n S(i, j)

    The p-value returned is calculated using a permutation test using
    :math:`hyppo.tools.perm_test`.

    The fast version of this test performs a multivariate independence test
    based on univariate test statistics :footcite:p:`hellerMultivariateTestsOfAssociation2016`.
    The univariate test statistic used is Hoeffding's independence test, derived as follows
    :footcite:p:`sasHoeffdingDependenceCoefficient`:

    Let :math:`x` and :math:`y` be :math:`(n, p)` samples of random variables
    :math:`X` and :math:`Y`. A center point - the center of mass of points in 'X' and 'Y'
    - is chosen. For every sample :math:`i`, calculate the distances from the center point
    in :math:`x` and :math:`y` and denote this as :math:`d_x(x_i)`
    and :math:`d_y(y_i)`. This will create a 1D collection of distances for each
    sample group.

    From these distances, we can calculate the Hoeffding's dependence score between
    the two groups - denoted as :math:`D` - using,

    .. math::

        D &= \frac{(n-2) (n-3) D_{1} + D_{2} - 2(n-2) D_{3}}
                 {n (n-1) (n-2) (n-3) (n-4)}

        D_{1} &= \sum_{i} (Q_{i}-1) (Q_{i}-2)

        D_{2} &= \sum_{i} (R_{i} - 1) (R_{i} - 2) (S_{i} - 1) (S_{i} - 2)

        D_{3} &= \sum_{i} {R_{i} - 2} (S_{i} - 2) (Q_{i}-1)

    where :math:`R_{i}` is the rank of :math:`x_{i}`,
    :math:`D_{i}` is the rank of :math:`y_{i}`,
    :math:`Q_{i}` is the bivariate rank = 1 plus the number of points with both x and y
    values less than the :math:`i`-th point.

    :math:`D` is notably sensitive to ties and gets smaller the more pairs of variables with identical values.
    If there are no ties in the data,D ranges between -0.5 and 1, with 1 indicating complete dependence.
    :footcite:p:`sasHoeffdingDependenceCoefficient`

    The p-value returned is calculated using a permutation test using
    :meth:`hyppo.tools.perm_test`.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, compute_distance="euclidean", **kwargs):
        self.is_distance = False
        if not compute_distance:
            self.is_distance = True
        self.auto = False
        IndependenceTest.__init__(self, compute_distance=compute_distance, **kwargs)

    def statistic(self, x, y):
        r"""
        Helper function that calculates the HHG test statistic.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions.
            Alternatively, ``x`` and ``y`` can be distance matrices,
            where the shapes must both be ``(n, n)``.
            For fast version, ``x`` and ``y`` can be 1D collections of distances
            from a chosen center point, where the shapes must be ``(n,1)`` or ``(n-1,1)``
            depending on choice of center point.

        Returns
        -------
        stat : float
            The computed HHG statistic.
        """
        distx = x
        disty = y

        if not self.is_distance:
            distx, disty = compute_dist(
                x, y, metric=self.compute_distance, **self.kwargs
            )

        if self.auto:
            if not self.is_distance:
                distx, disty = _centerpoint_dist(
                    x, y, metric=self.compute_distance, **self.kwargs
                )
            stat = hoeffdings(distx, disty)
        else:
            S = _pearson_stat(distx, disty)
            mask = np.ones(S.shape, dtype=bool)
            np.fill_diagonal(mask, 0)
            stat = np.sum(S[mask])
        self.stat = stat

        return stat

    def test(self, x, y, reps=1000, workers=1, auto=False, random_state=None):
        r"""
        Calculates the HHG test statistic and p-value.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions.
            Alternatively, ``x`` and ``y`` can be distance matrices,
            where the shapes must both be ``(n, n)``.
            For fast version, ``x`` and ``y`` can be 1D collections of distances
            from a chosen center point, where the shapes must be ``(n,1)`` or ``(n-1,1)``
            depending on choice of center point.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.
        auto : boolean, default: False
            Automatically use fast approximation of HHG test. :class:`hyppo.tools.perm_test`
            will still be run.

        Returns
        -------
        stat : float
            The computed HHG statistic.
        pvalue : float
            The computed HHG p-value.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.independence import HHG
        >>> x = np.arange(7)
        >>> y = x
        >>> stat, pvalue = HHG().test(x, y)
        >>> '%.1f, %.2f' % (stat, pvalue)
        '160.0, 0.00'

        In addition, the inputs can be distance matrices. Using this is the,
        same as before, except the ``compute_distance`` parameter must be set
        to ``None``.

        >>> import numpy as np
        >>> from hyppo.independence import HHG
        >>> x = np.ones((10, 10)) - np.identity(10)
        >>> y = 2 * x
        >>> hhg = HHG(compute_distance=None)
        >>> stat, pvalue = hhg.test(x, y)
        >>> '%.1f, %.2f' % (stat, pvalue)
        '0.0, 1.00'
        """
        check_input = _CheckInputs(x, y, reps=reps)
        x, y = check_input()
        self.auto = auto

        # Fast HHG Test
        if self.auto:
            distx, disty = _centerpoint_dist(
                x, y, metric=self.compute_distance, **self.kwargs
            )
            self.is_distance = True
            stat, pvalue = super(HHG, self).test(
                distx, disty, reps, workers, is_distsim=False
            )

        else:
            x, y = compute_dist(x, y, metric=self.compute_distance, **self.kwargs)
            self.is_distance = True
            stat, pvalue = super(HHG, self).test(x, y, reps, workers)

        return IndependenceTestOutput(stat, pvalue)


@jit(nopython=True, cache=True)
def _pearson_stat(distx, disty):  # pragma: no cover
    """Calculate the Pearson chi square stats"""

    n = distx.shape[0]
    S = np.zeros((n, n))

    # iterate over all samples in the distance matrix
    for i in range(n):
        for j in range(n):
            if i != j:
                a = distx[i, :] <= distx[i, j]
                b = disty[i, :] <= disty[i, j]

                t11 = np.sum(a * b) - 2
                t12 = np.sum(a * (1 - b))
                t21 = np.sum((1 - a) * b)
                t22 = np.sum((1 - a) * (1 - b))

                denom = (t11 + t12) * (t21 + t22) * (t11 + t21) * (t12 + t22)
                if denom > 0:
                    S[i, j] = ((n - 2) * (t12 * t21 - t11 * t22) ** 2) / denom

    return S


def hoeffdings(x, y):
    """For fast HHG, calculates the Hoeffding's dependence statistic"""
    R = rankdata(x)
    S = rankdata(y)

    # core processing
    N = x.shape
    D = _hoeffdings_d_calc(R, S, N)
    return D


@jit(nopython=True, cache=True)
def _hoeffdings_d_calc(R, S, N):  # pragma: no cover
    Q = np.ones(N[0])
    for i in range(0, N[0]):
        Q[i] = Q[i] + np.sum(np.bitwise_and(R < R[i], S < S[i]))
        Q[i] = Q[i] + 1 / 4 * (np.sum(np.bitwise_and(R == R[i], S == S[i])) - 1)
        Q[i] = Q[i] + 1 / 2 * (np.sum(np.bitwise_and(R == R[i], S < S[i])))
        Q[i] = Q[i] + 1 / 2 * (np.sum(np.bitwise_and(R < R[i], S == S[i])))

    D1 = np.sum(np.multiply((Q - 1), (Q - 2)))
    D2 = np.sum(
        np.multiply(np.multiply((R - 1), (R - 2)), np.multiply((S - 1), (S - 2)))
    )
    D3 = np.sum(np.multiply(np.multiply((R - 2), (S - 2)), (Q - 1)))

    D = (
        30
        * ((N[0] - 2) * (N[0] - 3) * D1 + D2 - 2 * (N[0] - 2) * D3)
        / (N[0] * (N[0] - 1) * (N[0] - 2) * (N[0] - 3) * (N[0] - 4))
    )
    return D


def _centerpoint_dist(x, y, metric, **kwargs):
    """Calculate the distance of x and y from center of mass"""
    pointer = (np.mean(x, axis=0), np.mean(y, axis=0))
    zx, zy = pointer
    zx = np.array(zx).reshape(1, -1)
    zy = np.array(zy).reshape(1, -1)
    xin = np.concatenate((zx, x))
    yin = np.concatenate((zy, y))
    distx, disty = compute_dist(xin, yin, metric=metric, **kwargs)
    # take first row of distance matrix = distance of sample points from center point
    distx = np.delete(distx[0], 0)
    distx = distx.reshape(-1, 1)
    disty = np.delete(disty[0], 0)
    disty = disty.reshape(-1, 1)

    return distx, disty



from __future__ import print_function, division, unicode_literals, absolute_import

from builtins import int

import autograd.numpy as np
from sklearn.metrics.pairwise import euclidean_distances
from numpy.random import default_rng


def constrain(val, min_val, max_val):
    return min(max_val, max(min_val, val))


def meddistance(X, subsample=None, mean_on_fail=True):
    r"""
    Compute the median of pairwise distances (not distance squared) of points
    in the matrix.  Useful as a heuristic for setting Gaussian kernel's width.

    Parameters
    ----------
    X : n x d numpy array
    mean_on_fail: True/False. If True, use the mean when the median distance is 0.
        This can happen especially, when the data are discrete e.g., 0/1, and
        there are more slightly more 0 than 1. In this case, the m

    Returns
    -------
    median distance

    From: https://github.com/wittawatj/fsic-test
    """
    if subsample is None:
        D = euclidean_distances(X, X)
        Itri = np.tril_indices(D.shape[0], -1)
        Tri = D[Itri]
        med = np.median(Tri)
        if med <= 0:
            # use the mean
            return np.mean(Tri)
        return med

    else:
        assert subsample > 0
        rand_state = np.random.get_state()
        rng = default_rng(9827)
        n = X.shape[0]
        ind = rng.choice(n, min(subsample, n), replace=False)
        np.random.set_state(rand_state)
        # recursion just one
        return meddistance(X[ind, :], None, mean_on_fail)


def fit_gaussian_draw(X, J, seed=28, reg=1e-7, eig_pow=1.0):
    r"""
    Fit a multivariate normal to the data X (n x d) and draw J points
    from the fit.

    Parameters
    ----------
    reg : regularizer to use with the covariance matrix
    eig_pow : raise eigenvalues of the covariance matrix to this power to construct
        a new covariance matrix before drawing samples. Useful to shrink the spread
        of the variance.

    From: https://github.com/wittawatj/fsic-test
    """
    rng = default_rng(seed)
    d = X.shape[1]
    mean_x = np.mean(X, 0)
    cov_x = np.cov(X.T)
    if d == 1:
        cov_x = np.array([[cov_x]])
    [evals, evecs] = np.linalg.eig(cov_x)
    evals = np.maximum(0, np.real(evals))
    assert np.all(np.isfinite(evals))
    evecs = np.real(evecs)
    shrunk_cov = evecs.dot(np.diag(evals**eig_pow)).dot(evecs.T) + reg * np.eye(d)
    V = rng.multivariate_normal(mean_x, shrunk_cov, J)
    return V


def outer_rows(X, Y):
    r"""
    Compute the outer product of each row in X, and Y.

    Parameters
    ----------
    X : n x dx numpy array
    Y : n x dy numpy array

    Returns
    -------
    Return an n x dx x dy numpy array.
    """
    return np.einsum("ij,ik->ijk", X, Y)



import random
from typing import NamedTuple

import numpy as np
from numba import jit

from ..tools import perm_test
from .base import KSampleTest


class FRTestOutput(NamedTuple):
    stat: float
    pvalue: float
    uncor_stat: dict


class FriedmanRafsky(KSampleTest):
    r"""
    Friedman-Rafksy (FR) test statistic and p-value.

    This is a multivariate extension of the Wald-Wolfowitz
    runs test for randomness. The normal concept of a 'run'
    is replaced by a minimum spanning tree (MST) calculated between
    the points in respective data sets with edge weights defined
    as the Euclidean distance between two such points. After MST
    has been determined, all edges such that both corresponding
    nodes do not belong to the same class are severed and the
    number of independent resulting trees is counted. This test is
    consistent against similar tests.

    Notes
    -----
    The statistic can be derived as follows
    :footcite:p:`friedmanMultivariateGeneralizationsoftheWaldWolfowitzandSmirnovTwoSampleTests1979`

    Let :math:`x` be a combined sample of :math:`(n, p)` and :math:`(m, p)`
    samples of random variables :math:`X` and let :math:`y` be a :math:`(n+m, 1)`
    array of labels :math:`Y`. We can then create a graph such that each point in
    :math:`X` is connected to each other point in :math:`X` by an edge weighted by
    the euclidean distance inbetween those points. The minimum spanning tree is then
    calculated and all edges such that the labels in :math:`Y` are not from the same
    class are removed. The number of independent graphs is then summed to determine
    the uncorrected statistic for the test.

    The p-value and null distribution for the corrected statistic are calculated via
    a permutation test using :meth:`hyppo.tools.perm_test`.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, **kwargs):

        KSampleTest.__init__(self, **kwargs)

    def statistic(self, x, y):
        r"""
        Helper function that calculates the Friedman Rafksy test statistic.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            rows. That is, the shapes must be ``(n, p)`` and ``(n, 1)`` where
            `n` is the number of combined samples and `p` is the number of
            dimensions. ``y`` is the array of labels corresponding to the two
            samples, respectively.

        Returns
        -------
        stat : float
            The computed (uncorrected) Friedman Rafsky statistic. A value between
            ``2`` and ``n``.
        """
        x = np.transpose(x)
        labels = np.transpose(y)

        MST_connections = MST(x, labels)
        stat = _num_runs(labels, MST_connections)

        return stat

    def test(
        self,
        x,
        y,
        reps=1000,
        workers=1,
        random_state=None,
    ):
        r"""
        Calculates the Friedman Rafsky test statistic and p-value.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            rows. That is, the shapes must be ``(n, p)`` and ``(n, 1)`` where
            `n` is the number of combined samples and `p` is the number of
            dimensions. ``y`` is the array of labels corresponding to the two
            samples, respectively.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.
        random_state : int, default: None
            The random_state for permutation testing to be fixed for
            reproducibility.

        Returns
        -------
        stat : float
            The computed (corrected) Friedman Rafsky statistic.
        pvalue : float
            The computed Friedman Rafsky p-value.
        uncor_stat : float
            The computed (uncorrected) Friedman Rafsky statistic.
        """
        uncor_stat, pvalue, null_dist = perm_test(
            self.statistic,
            x,
            y,
            reps=reps,
            workers=workers,
            is_distsim=False,
            random_state=random_state,
        )
        self.uncor_stat = uncor_stat
        stat = (uncor_stat - np.mean(null_dist)) / np.std(null_dist)
        self.stat = stat

        return FRTestOutput(stat, pvalue, uncor_stat)


def _num_runs(labels, MST_connections):
    r"""
    Helper function to determine number of independent
    'runs' from MST connections.

    Parameters
    ----------
    labels : ndarry of float
        Lables corresponding to respective classes of samples.
    MST_connections: list of int
        List containing pairs of points connected in final MST.

    Returns
    -------
    run_count : int
        Number of runs after severing all such edges with nodes of
        differing class labels.
    """
    run_count = 1

    for x in MST_connections:
        if labels[x[0]] != labels[x[1]]:
            run_count += 1

    return run_count


@jit(nopython=True, cache=True)  # pragma: no cover
def prim(weight_mat, labels):
    r"""
    Helper function to read weighted matrix input and compute minimum
    spanning tree via Prim's algorithm.

    Parameters
    ----------
    weight_mat : ndarry of float
        Weighted connection matrix.
    labels : ndarry of int
        Lables corresponding to respective classes of samples.

    Returns
    -------
    MST_connections : list of int
        List of pairs of nodes connected in final MST.
    """
    INF = 9999999
    V = len(labels)
    selected = np.zeros(len(labels))
    no_edge = 0
    selected[0] = True
    MST_connections = []

    while no_edge < V - 1:
        minimum = INF
        x = 0
        y = 0
        for i in range(V):
            if selected[i]:
                for j in range(V):
                    if (not selected[j]) and weight_mat[i][j]:
                        # not in selected and there is an edge
                        if minimum > weight_mat[i][j]:
                            minimum = weight_mat[i][j]
                            x = i
                            y = j
        MST_connections.append([x, y])
        selected[y] = True
        no_edge += 1

    return MST_connections


@jit(nopython=True, cache=True)
def MST(x, labels):  # pragma: no cover
    r"""
    Helper function to read input data and calculate Euclidean distance
    between each possible pair of points before finding MST.

    Parameters
    ----------
    x : ndarry of float
        Dataset such that each column corresponds to a point of data.
    labels : ndarry of int
        Lables corresponding to respective classes of samples.

    Returns
    -------
    MST_connections : list
        List of pairs of nodes connected in final MST.
    """

    G = np.zeros((len(x[0]), len(x[0])))

    for i in range(len(x[0])):

        for j in range(i + 1, len(x[0])):
            weight = np.linalg.norm(x[:, i] - x[:, j])
            G[i][j] = weight
            G[j][i] = weight

    MST_connections = prim(G, labels)

    return MST_connections



from abc import ABC, abstractmethod
from typing import NamedTuple


class KSampleTestOutput(NamedTuple):
    stat: float
    pvalue: float


class KSampleTest(ABC):
    """
    A base class for a *k*-sample test.

    Parameters
    ----------
    compute_distance : str, callable, or None, default: "euclidean" or "gaussian"
        A function that computes the distance among the samples within each
        data matrix.
        Valid strings for ``compute_distance`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,

            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.

        Alternatively, this function computes the kernel similarity among the
        samples within each data matrix.
        Valid strings for ``compute_kernel`` are, as defined in
        :func:`sklearn.metrics.pairwise.pairwise_kernels`,

            [``"additive_chi2"``, ``"chi2"``, ``"linear"``, ``"poly"``,
            ``"polynomial"``, ``"rbf"``,
            ``"laplacian"``, ``"sigmoid"``, ``"cosine"``]

        Note ``"rbf"`` and ``"gaussian"`` are the same metric.
    bias : bool (default: False)
        Whether or not to use the biased or unbiased test statistics. Only
        applies to ``Dcorr`` and ``Hsic``.
    **kwargs
        Arbitrary keyword arguments for ``compute_distkern``.
    """

    def __init__(self, compute_distance=None, bias=False, **kwargs):
        # set statistic and p-value
        self.stat = None
        self.pvalue = None
        self.bias = bias
        self.compute_distance = compute_distance
        self.kwargs = kwargs

        super().__init__()

    @abstractmethod
    def statistic(self, *args):
        r"""
        Calulates the *k*-sample test statistic.

        Parameters
        ----------
        *args : ndarray of float
            Variable length input data matrices. All inputs must have the same
            number of dimensions. That is, the shapes must be `(n, p)` and
            `(m, p)`, ... where `n`, `m`, ... are the number of samples and `p` is
            the number of dimensions.

        Returns
        -------
        stat : float
            The computed *k*-Sample statistic.
        """

    @abstractmethod
    def test(self, *args, reps=1000, workers=1, random_state=None):
        r"""
        Calculates the *k*-sample test statistic and p-value.

        Parameters
        ----------
        *args : ndarray of float
            Variable length input data matrices. All inputs must have the same
            number of dimensions. That is, the shapes must be `(n, p)` and
            `(m, p)`, ... where `n`, `m`, ... are the number of samples and `p` is
            the number of dimensions.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.

        Returns
        -------
        stat : float
            The computed *k*-sample statistic.
        pvalue : float
            The computed *k*-sample p-value.
        """



from itertools import permutations

import numpy as np
import pytest
from numpy.testing import assert_raises

from .._utils import _CheckInputs


class TestErrorWarn:
    """Tests errors and warnings."""

    def test_error_notndarray(self):
        # raises error if x or y is not a ndarray
        x = np.arange(20)
        y = np.arange(20)
        z = [5] * 20

        for data in permutations((x, y, z)):
            assert_raises(TypeError, _CheckInputs(*data))

    def test_error_shape(self):
        # raises error if number of samples different (n)
        x = np.arange(100).reshape(25, 4)
        y = x.reshape(10, 10)
        z = x.reshape(20, 5)
        assert_raises(ValueError, _CheckInputs(x, y, z))

    def test_error_lowsamples(self):
        # raises error if samples are low (< 3)
        x = np.arange(3)
        y = np.arange(3)
        z = np.arange(3)
        assert_raises(ValueError, _CheckInputs(x, y, z))

    def test_error_nans(self):
        # raises error if inputs contain NaNs
        x = np.arange(20, dtype=float)
        x[0] = np.nan
        assert_raises(ValueError, _CheckInputs(x, x, x))

        y = np.arange(20)
        assert_raises(ValueError, _CheckInputs(x, y, y))

    @pytest.mark.parametrize(
        "reps", [-1, "1"]  # reps is negative  # reps is not integer
    )
    def test_error_reps(self, reps):
        # raises error if reps is negative
        x = np.arange(20)
        assert_raises(ValueError, _CheckInputs(x, x, x, reps=reps))

    def test_max_sims(self):
        # raises error if data contains more dimentions than allowed
        x = np.arange(20).reshape(10, 2)
        y = np.arange(10).reshape(10, 1)
        z = y

        assert_raises(ValueError, _CheckInputs(x, y, z, max_dims=1))

    def test_constant_input(self):
        # raises error if z is constant
        x = np.arange(20).reshape(-1, 1)
        y = np.arange(20).reshape(-1, 1)
        z = np.ones(20).reshape(-1, 1)

        assert_raises(ValueError, _CheckInputs(x, y, z, ignore_z_var=False))

        try:
            _CheckInputs(x, y, z, ignore_z_var=True)
        except Exception as exc:
            assert False, f"'_CheckInputs' with 'ignore_z_var=True' with constant z input raised an exception {exc}"


import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from hyppo.tools import rot_ksamp
from hyppo.conditional import FCIT


class TestFCIT:
    @pytest.mark.parametrize(
        "n, obs_stat, obs_pvalue",
        [
            (2000, 11.677197, 3.8168e-06),
            (1000, 7.733, 5.378401185056561e-05),
        ],
    )
    def test_linear_oned(self, n, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y = rot_ksamp("linear", n, 1, k=2)
        np.random.seed(123456789)
        stat, pvalue = FCIT().test(x, y)

        assert_almost_equal(stat, obs_stat, decimal=-1)
        assert_almost_equal(pvalue, obs_pvalue, decimal=4)

    @pytest.mark.parametrize(
        "dim, n, obs_stat, obs_pvalue",
        [(1, 100000, -0.16024, 0.56139), (2, 100000, -4.59882, 0.99876)],
    )
    def test_null(self, dim, n, obs_stat, obs_pvalue):
        np.random.seed(12)
        z1 = np.random.multivariate_normal(
            mean=np.zeros(dim), cov=np.eye(dim), size=(n)
        )
        A1 = np.random.normal(loc=0, scale=1, size=dim * dim).reshape(dim, dim)
        B1 = np.random.normal(loc=0, scale=1, size=dim * dim).reshape(dim, dim)
        x1 = (
            A1 @ z1.T
            + np.random.multivariate_normal(
                mean=np.zeros(dim), cov=np.eye(dim), size=(n)
            ).T
        )
        y1 = (
            B1 @ z1.T
            + np.random.multivariate_normal(
                mean=np.zeros(dim), cov=np.eye(dim), size=(n)
            ).T
        )

        np.random.seed(122)
        stat, pvalue = FCIT().test(x1.T, y1.T, z1)

        assert_almost_equal(pvalue, obs_pvalue, decimal=4)
        assert_almost_equal(stat, obs_stat, decimal=4)

    @pytest.mark.parametrize(
        "dim, n, obs_stat, obs_pvalue",
        [
            (1, 100000, 89.271754, 2.91447597e-12),
            (2, 100000, 161.35165, 4.63412957e-14),
        ],
    )
    def test_alternative(self, dim, n, obs_stat, obs_pvalue):
        np.random.seed(12)
        z2 = np.random.multivariate_normal(
            mean=np.zeros(dim), cov=np.eye(dim), size=(n)
        )

        A2 = np.random.normal(loc=0, scale=1, size=dim * dim).reshape(dim, dim)
        B2 = np.random.normal(loc=0, scale=1, size=dim * dim).reshape(dim, dim)

        x2 = (
            A2 @ z2.T
            + np.random.multivariate_normal(
                mean=np.zeros(dim), cov=np.eye(dim), size=(n)
            ).T
        )
        y2 = (
            B2 @ x2
            + np.random.multivariate_normal(
                mean=np.zeros(dim), cov=np.eye(dim), size=(n)
            ).T
        )

        np.random.seed(122)
        stat, pvalue = FCIT().test(x2.T, y2.T, z2)

        assert_almost_equal(pvalue, obs_pvalue, decimal=12)
        assert_almost_equal(stat, obs_stat, decimal=4)



import numpy as np
import pytest
from numpy.testing import assert_almost_equal
from sklearn.metrics.pairwise import pairwise_kernels

from ...tools import linear, power
from .. import Hsic


class TestHsicStat:
    @pytest.mark.parametrize("n, obs_stat", [(100, 1.0), (200, 1.0)])
    @pytest.mark.parametrize("obs_pvalue", [1 / 1000])
    def test_linear_oned(self, n, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y = linear(n, 1)
        stat, pvalue = Hsic().test(x, y)

        assert_almost_equal(stat, obs_stat, decimal=2)
        assert_almost_equal(pvalue, obs_pvalue, decimal=2)

    def test_kernstat(self):
        np.random.seed(123456789)
        x, y = linear(100, 1)
        kernx = pairwise_kernels(x, x)
        kerny = pairwise_kernels(y, y)
        stat, pvalue = Hsic(compute_kernel=None).test(kernx, kerny, auto=False)

        assert_almost_equal(stat, 1.0, decimal=2)
        assert_almost_equal(pvalue, 1 / 1000, decimal=2)

    @pytest.mark.parametrize("n", [100, 200])
    def test_rep(self, n):
        x, y = linear(n, 1)
        stat, pvalue = Hsic().test(x, y, random_state=2)
        stat2, pvalue2 = Hsic().test(x, y, random_state=2)

        assert stat == stat2
        assert pvalue == pvalue2


class TestHsicTypeIError:
    def test_oned(self):
        np.random.seed(123456789)
        est_power = power(
            "Hsic",
            sim_type="indep",
            sim="multimodal_independence",
            n=200,
            p=1,
            alpha=0.05,
        )

        assert_almost_equal(est_power, 0.05, decimal=2)

    def test_oned_fast(self):
        np.random.seed(123456789)
        est_power = power(
            "Hsic",
            sim_type="indep",
            sim="multimodal_independence",
            n=500,
            p=1,
            alpha=0.05,
            auto=True,
        )

        assert_almost_equal(est_power, 0.05, decimal=2)



import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_raises, assert_warns

from ...tools import linear, power
from .. import FriedmanRafsky


class TestFriedmanRafskyStat:
    @pytest.mark.parametrize("n", [100])
    @pytest.mark.parametrize("num_runs", [48])
    @pytest.mark.parametrize("obs_stat", [-0.527])
    @pytest.mark.parametrize("obs_pvalue", [0.748])
    def test_linear_oned(self, n, num_runs, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y = linear(n, 1)
        num_rows, num_cols = x.shape
        y = np.random.choice([0, 1], num_rows, p=[0.5, 0.5])
        y = np.transpose(y)
        stat1, pvalue1, _ = FriedmanRafsky().test(x, y)
        stat2 = FriedmanRafsky().statistic(x, y)

        assert_almost_equal(stat1, obs_stat, decimal=2)
        assert_almost_equal(stat2, num_runs, decimal=2)
        assert_almost_equal(pvalue1, obs_pvalue, decimal=2)

    @pytest.mark.parametrize("n", [100, 200])
    def test_rep(self, n):
        x, y = linear(n, 1)
        num_rows, num_cols = x.shape
        y = np.random.choice([0, 1], num_rows, p=[0.5, 0.5])
        y = np.transpose(y)
        stat1, pvalue1, _ = FriedmanRafsky().test(x, y, random_state=2)
        stat2, pvalue2, _ = FriedmanRafsky().test(x, y, random_state=2)

        assert stat1 == stat2
        assert pvalue1 == pvalue2



from ..independence.dcorr import _dcov
from ..tools import compute_dist
from ._utils import _CheckInputs, k_sample_transform
from .base import KSampleTest, KSampleTestOutput
from .ksamp import KSample


class Energy(KSampleTest):
    r"""
    Energy test statistic and p-value.

    Energy is a powerful multivariate 2-sample test. It leverages distance matrix
    capabilities (similar to tests like distance correlation or Dcorr). In fact, Energy
    statistic is equivalent to our 2-sample formulation nonparametric MANOVA via
    independence testing, i.e. :class:`hyppo.ksample.KSample`,
    and to
    :class:`hyppo.independence.Dcorr`,
    :class:`hyppo.ksample.DISCO`,
    :class:`hyppo.independence.Hsic`, and
    :class:`hyppo.ksample.MMD`
    :footcite:p:`pandaNonparMANOVAIndependence2021`
    :footcite:p:`shenExactEquivalenceDistance2020`.

    Parameters
    ----------
    compute_distance : str, callable, or None, default: "euclidean"
        A function that computes the distance among the samples within each
        data matrix.
        Valid strings for ``compute_distance`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,

            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.

        Set to ``None`` or ``"precomputed"`` if ``x`` and ``y`` are already distance
        matrices. To call a custom function, either create the distance matrix
        before-hand or create a function of the form ``metric(x, **kwargs)``
        where ``x`` is the data matrix for which pairwise distances are
        calculated and ``**kwargs`` are extra arguements to send to your custom
        function.
    bias : bool, default: False
        Whether or not to use the biased or unbiased test statistics.
    **kwargs
        Arbitrary keyword arguments for ``compute_distance``.

    Notes
    -----
    Traditionally, the formulation for the 2-sample Energy statistic
    is as follows :footcite:p:`szekelyTestingEqualDistributions`:

    Define
    :math:`\{ u_i \stackrel{iid}{\sim} F_U,\ i = 1, ..., n \}` and
    :math:`\{ v_j \stackrel{iid}{\sim} F_V,\ j = 1, ..., m \}` as two groups
    of samples deriving from different distributions with the same
    dimensionality. If :math:`d(\cdot, \cdot)` is a distance metric (i.e. Euclidean)
    then,

    .. math::

        \mathrm{Energy}_{n, m}(\mathbf{u}, \mathbf{v}) = \frac{1}{n^2 m^2}
        \left( 2nm \sum_{i = 1}^n \sum_{j = 1}^m d(u_i, v_j) - m^2
        \sum_{i,j=1}^n d(u_i, u_j) - n^2 \sum_{i, j=1}^m d(v_i, v_j) \right)

    The implementation in the :class:`hyppo.ksample.KSample` class (using
    :class:`hyppo.independence.Dcorr` using 2 samples) is in
    fact equivalent to this implementation (for p-values) and statistics are
    equivalent up to a scaling factor :footcite:p:`pandaNonparMANOVAIndependence2021`.

    The p-value returned is calculated using a permutation test uses
    :meth:`hyppo.tools.perm_test`.
    The fast version of the test uses :meth:`hyppo.tools.chi2_approx`.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, compute_distance="euclidean", bias=False, **kwargs):
        KSampleTest.__init__(
            self, compute_distance=compute_distance, bias=bias, **kwargs
        )

    def statistic(self, x, y):
        r"""
        Calulates the Energy test statistic.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            dimensions. That is, the shapes must be ``(n, p)`` and ``(m, p)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions.

        Returns
        -------
        stat : float
            The computed Energy statistic.
        """
        distx = x
        disty = y
        n = x.shape[0]
        m = y.shape[0]

        x, y = k_sample_transform([x, y])
        distx, disty = compute_dist(x, y, metric=self.compute_distance, **self.kwargs)

        # exact equivalence transformation Dcorr and Energy
        stat = (
            _dcov(distx, disty, bias=self.bias, only_dcov=True)
            * ((n + m) ** 4)
            / (2 * (n**2) * (m**2))
        )
        self.stat = stat

        return stat

    def test(self, x, y, reps=1000, workers=1, auto=True, random_state=None):
        r"""
        Calculates the Energy test statistic and p-value.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            dimensions. That is, the shapes must be ``(n, p)`` and ``(m, p)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.
        auto : bool, default: True
            Automatically uses fast approximation when `n` and size of array
            is greater than 20. If ``True``, and sample size is greater than 20, then
            :class:`hyppo.tools.chi2_approx` will be run. Parameters ``reps`` and
            ``workers`` are
            irrelevant in this case. Otherwise, :class:`hyppo.tools.perm_test` will be
            run.

        Returns
        -------
        stat : float
            The computed Energy statistic.
        pvalue : float
            The computed Energy p-value.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.ksample import Energy
        >>> x = np.arange(7)
        >>> y = x
        >>> stat, pvalue = Energy().test(x, y)
        >>> '%.3f, %.1f' % (stat, pvalue)
        '-0.895, 1.0'
        """
        check_input = _CheckInputs(
            inputs=[x, y],
            indep_test="dcorr",
        )
        x, y = check_input()

        # observed statistic
        stat = self.statistic(x, y)

        # since stat transformation is invariant under permutation, 2-sample Dcorr
        # pvalue is identical to Energy
        _, pvalue = KSample(
            indep_test="Dcorr",
            compute_distkern=self.compute_distance,
            bias=self.bias,
            **self.kwargs
        ).test(x, y, reps=reps, workers=workers, auto=auto, random_state=random_state)

        return KSampleTestOutput(stat, pvalue)



import numpy as np
from numpy import testing
from past.utils import old_div
from scipy.linalg.misc import norm

from ..datasource import DSNormal, DSIsotropicNormal
from ..density import IsotropicNormal, Normal
import scipy.stats as stats
from numpy.random import default_rng

import pytest


class TestIsotropicNormal:
    @pytest.mark.parametrize("n", [7])
    @pytest.mark.parametrize("d", [3, 1])
    def test_log_den(self, n, d):
        rng = default_rng(16)
        variance = 1.1
        mean = rng.standard_normal(size=d)
        X = rng.random(size=(n, d)) + 1

        isonorm = IsotropicNormal(mean, variance)
        ds_isonorm = DSIsotropicNormal(mean, variance)
        isonorm.log_normalized_den(X)
        isonorm.dim()
        log_dens = isonorm.log_den(X)
        my_log_dens = -np.sum((X - mean) ** 2, 1) / (2.0 * variance)

        # check correctness
        testing.assert_almost_equal(log_dens, my_log_dens)

    @pytest.mark.parametrize("n", [8])
    @pytest.mark.parametrize("d", [4, 1])
    def test_grad_log(self, n, d):
        rng = default_rng(17)
        variance = 1.2
        mean = rng.standard_normal(size=d) + 1
        X = rng.random(size=(n, d)) - 2

        isonorm = IsotropicNormal(mean, variance)
        grad_log = isonorm.grad_log(X)
        my_grad_log = -(X - mean) / variance

        # check correctness
        testing.assert_almost_equal(grad_log, my_grad_log)


class TestNormal:
    @pytest.mark.parametrize("n", [2])
    @pytest.mark.parametrize("d", [2])
    def test_log_den(self, n, d):
        rng = default_rng(16)
        cov = np.array([[1.1, 1.2], [1.1, 1.2]])
        mean = rng.standard_normal(size=(n, d))
        X = rng.random(size=(n, d)) + 1

        test_mean = np.ones(2)
        test_cov = np.array([[1.1, 1.2], [1.1, 1.2]])

        norm = Normal(mean, cov)
        log_dens = norm.log_den(X)
        E, V = np.linalg.eigh(cov)
        prec = np.dot(np.dot(V, np.diag(old_div(1.0, E))), V.T)
        X0 = X - mean
        X0prec = np.dot(X0, prec)
        my_log_dens = old_div(-np.sum(X0prec * X0, 1), 2.0)

        ds_norm = DSNormal(test_mean, cov)
        ds_norm.sample(n=10)
        norm.get_datasource()
        norm.dim()

        # check correctness
        testing.assert_almost_equal(log_dens, my_log_dens)



import numpy as np


class _CheckInputs:
    """Check if additional arguments are correct."""

    def __init__(self, n, p):
        self.n = n
        self.p = p

    def __call__(self, **kwargs):
        if type(self.n) is not int or type(self.p) is not int:
            raise ValueError(
                "Expected n and p of type int, got {} and {}".format(
                    type(self.n), type(self.p)
                )
            )

        if self.n < 5:
            raise ValueError("n must be >= 5, got {}".format(self.n))

        if self.p < 1:
            raise ValueError("p must be >= 1, got {}".format(self.p))

        for key, value in kwargs.items():
            if value[1] is float and type(value[0]) is int:
                continue
            if type(value[0]) is not value[1]:
                raise ValueError(
                    "Expected {} type {} got {}".format(key, value[1], type(value[0]))
                )


def _gen_coeffs(p):
    """Calculates coefficients polynomials."""
    return np.array([1 / (i + 1) for i in range(p)]).reshape(-1, 1)


def _random_uniform(n, p, low=-1, high=1):
    """Generate random uniform data."""
    return np.array(np.random.uniform(low, high, size=(n, p)))


def _calc_eps(n):
    """Calculate noise."""
    return np.random.normal(0, 1, size=(n, 1))


def linear(n, p, noise=False, low=-1, high=1):
    r"""
    Linear simulation.

    Linear :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}`:

    .. math::

        X &\sim \mathcal{U}(-1, 1)^p \\
        Y &= w^T X + \kappa \epsilon

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.
    low : float, default: -1
        The lower limit of the uniform distribution simulated from.
    high : float, default: 1
        The upper limit of the uniform distribution simulated from.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, 1)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    extra_args = {
        "noise": (noise, bool),
        "low": (low, float),
        "high": (high, float),
    }
    check_in = _CheckInputs(n, p)
    check_in(**extra_args)

    x = _random_uniform(n, p, low, high)
    coeffs = _gen_coeffs(p)
    eps = _calc_eps(n)
    y = x @ coeffs + 1 * noise * eps

    return x, y


def exponential(n, p, noise=False, low=0, high=3):
    r"""
    Exponential simulation.

    Exponential :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}`:

    .. math::

        X &\sim \mathcal{U}(0, 3)^p \\
        Y &= \exp (w^T X) + 10 \kappa \epsilon

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.
    low : float, default: 0
        The lower limit of the uniform distribution simulated from.
    high : float, default: 3
        The upper limit of the uniform distribution simulated from.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, 1)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    extra_args = {
        "noise": (noise, bool),
        "low": (low, float),
        "high": (high, float),
    }
    check_in = _CheckInputs(n, p)
    check_in(**extra_args)

    x = _random_uniform(n, p, low, high)
    coeffs = _gen_coeffs(p)
    eps = _calc_eps(n)
    y = np.exp(x @ coeffs) + 10 * noise * eps

    return x, y


def cubic(n, p, noise=False, low=-1, high=1, cubs=[-12, 48, 128], scale=1 / 3):
    r"""
    Cubic simulation.

    Cubic :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}`:

    .. math::

        X &\sim \mathcal{U}(-1, 1)^p \\
        Y &= 128 \left( w^T X - \frac{1}{3} \right)^3
             + 48 \left( w^T X - \frac{1}{3} \right)^2
             - 12 \left( w^T X - \frac{1}{3} \right)
             + 80 \kappa \epsilon

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.
    low : float, default: -1
        The lower limit of the uniform distribution simulated from.
    high : float, default: 1
        The upper limit of the uniform distribution simulated from.
    cubs : list of ints, default: [-12, 48, 128]
        Coefficients of the cubic function where each value corresponds to the
        order of the cubic polynomial.
    scale : float, default: 1/3
        Scaling center of the cubic.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, 1)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    extra_args = {
        "noise": (noise, bool),
        "low": (low, float),
        "high": (high, float),
        "cubs": (cubs, list),
        "scale": (scale, float),
    }
    check_in = _CheckInputs(n, p)
    check_in(**extra_args)

    x = _random_uniform(n, p, low, high)
    coeffs = _gen_coeffs(p)
    eps = _calc_eps(n)

    x_coeffs = x @ coeffs - scale
    y = (
        cubs[2] * x_coeffs**3
        + cubs[1] * x_coeffs**2
        + cubs[0] * x_coeffs**3
        + 80 * noise * eps
    )

    return x, y


def joint_normal(n, p, noise=False):
    r"""
    Joint Normal simulation.

    Joint Normal :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}^p`: Let
    :math:`\rho = \frac{1}{2} p`, :math:`I_p` be the identity matrix of size
    :math:`p \times p`, :math:`J_p` be the matrix of ones of size
    :math:`p \times p` and
    :math:`\Sigma = \begin{bmatrix} I_p & \rho J_p \\ \rho J_p & (1 + 0.5\kappa) I_p \end{bmatrix}`.
    Then,

    .. math::

        (X, Y) \sim \mathcal{N}(0, \Sigma)

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, p)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    if p > 10:
        raise ValueError("Covariance matrix for p > 10 is not positive semi-definite")

    extra_args = {
        "noise": (noise, bool),
    }
    check_in = _CheckInputs(n, p)
    check_in(**extra_args)

    rho = 1 / (2 * p)
    cov1 = np.concatenate((np.identity(p), rho * np.ones((p, p))), axis=1)
    cov2 = np.concatenate((rho * np.ones((p, p)), np.identity(p)), axis=1)
    covT = np.concatenate((cov1.T, cov2.T), axis=1)

    eps = _calc_eps(n)
    x = np.random.multivariate_normal(np.zeros(2 * p), covT, n)
    y = x[:, p : 2 * p] + 0.5 * noise * eps
    x = x[:, :p]

    return x, y


def step(n, p, noise=False, low=-1, high=1):
    r"""
    Step simulation.

    Step :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}`:

    .. math::

        X &\sim \mathcal{U}(-1, 1)^p \\
        Y &= \mathbb{1}_{w^T X > 0} + \epsilon

    where :math:`\mathbb{1}` is the indicator function.

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.
    low : float, default: -1
        The lower limit of the uniform distribution simulated from.
    high : float, default: 1
        The upper limit of the uniform distribution simulated from.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, 1)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    extra_args = {
        "noise": (noise, bool),
        "low": (low, float),
        "high": (high, float),
    }
    check_in = _CheckInputs(n, p)
    check_in(**extra_args)

    if p > 1:
        noise = True
    x = _random_uniform(n, p, low, high)
    coeffs = _gen_coeffs(p)
    eps = _calc_eps(n)

    x_coeff = ((x @ coeffs) > 0) * 1
    y = x_coeff + noise * eps

    return x, y


def quadratic(n, p, noise=False, low=-1, high=1):
    r"""
    Quadratic simulation.

    Quadratic :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}`:

    .. math::

        X &\sim \mathcal{U}(-1, 1)^p \\
        Y &= (w^T X)^2 + 0.5 \kappa \epsilon

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.
    low : float, default: -1
        The lower limit of the uniform distribution simulated from.
    high : float, default: 1
        The upper limit of the uniform distribution simulated from.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, 1)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    extra_args = {
        "noise": (noise, bool),
        "low": (low, float),
        "high": (high, float),
    }
    check_in = _CheckInputs(n, p)
    check_in(**extra_args)

    x = _random_uniform(n, p, low, high)
    coeffs = _gen_coeffs(p)
    eps = _calc_eps(n)

    x_coeffs = x @ coeffs
    y = x_coeffs**2 + 0.5 * noise * eps

    return x, y


def w_shaped(n, p, noise=False, low=-1, high=1):
    r"""
    W-Shaped simulation.

    W-Shaped :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}`:
    :math:`\mathcal{U}(-1, 1)^p`,

    .. math::

        X &\sim \mathcal{U}(-1, 1)^p \\
        Y &= \left[ \left( (w^T X)^2 - \frac{1}{2} \right)^2
                            + \frac{w^T U}{500} \right] + 0.5 \kappa \epsilon

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.
    low : float, default: -1
        The lower limit of the uniform distribution simulated from.
    high : float, default: 1
        The upper limit of the uniform distribution simulated from.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, 1)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    extra_args = {
        "noise": (noise, bool),
        "low": (low, float),
        "high": (high, float),
    }
    check_in = _CheckInputs(n, p)
    check_in(**extra_args)

    x = _random_uniform(n, p, low, high)
    u = _random_uniform(n, p, 0, 1)
    coeffs = _gen_coeffs(p)
    eps = _calc_eps(n)

    x_coeffs = x @ coeffs
    u_coeffs = u @ coeffs
    y = 4 * ((x_coeffs**2 - 0.5) ** 2 + u_coeffs / 500) + 0.5 * noise * eps

    return x, y


def spiral(n, p, noise=False, low=0, high=5):
    r"""
    Spiral simulation.

    Spiral :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}`:
    :math:`U \sim \mathcal{U}(0, 5)`, :math:`\epsilon \sim \mathcal{N}(0, 1)`

    .. math::

        X_{|d|} &= U \sin(\pi U) \cos^d(\pi U)\ \mathrm{for}\ d = 1,...,p-1 \\
        X_{|p|} &= U \cos^p(\pi U) \\
        Y &= U \sin(\pi U) + 0.4 p \epsilon

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.
    low : float, default: 0
        The lower limit of the uniform distribution simulated from.
    high : float, default: 5
        The upper limit of the uniform distribution simulated from.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, 1)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    extra_args = {
        "noise": (noise, bool),
        "low": (low, float),
        "high": (high, float),
    }
    check_in = _CheckInputs(n, p)
    check_in(**extra_args)

    if p > 1:
        noise = True
    rx = _random_uniform(n, p=1, low=low, high=high)
    ry = rx
    rx = np.repeat(rx, p, axis=1)
    z = rx
    x = np.zeros((n, p))
    x[:, 0] = np.cos(z[:, 0] * np.pi)
    for i in range(p - 1):
        x[:, i + 1] = np.multiply(x[:, i], np.cos(z[:, i + 1] * np.pi))
        x[:, i] = np.multiply(x[:, i], np.sin(z[:, i + 1] * np.pi))
    x = np.multiply(rx, x)
    y = np.multiply(ry, np.sin(z[:, 0].reshape(-1, 1) * np.pi))

    eps = _calc_eps(n)
    y = y + 0.4 * p * noise * eps

    return x, y


def uncorrelated_bernoulli(n, p, noise=False, prob=0.5):
    r"""
    Uncorrelated Bernoulli simulation.

    Uncorrelated Bernoulli :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}`:
    :math:`U \sim \mathcal{B}(0.5)`, :math:`\epsilon_1 \sim \mathcal{N}(0, I_p)`,
    :math:`\epsilon_2 \sim \mathcal{N}(0, 1)`,

    .. math::

        X &= \mathcal{B}(0.5)^p + 0.5 \epsilon_1 \\
        Y &= (2U - 1) w^T X + 0.5 \epsilon_2

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.
    prob : float, default: 0.5
        The probability of the bernoulli distribution simulated from.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, 1)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    extra_args = {
        "noise": (noise, bool),
        "prob": (prob, float),
    }
    check_in = _CheckInputs(n, p)
    check_in(**extra_args)

    binom = np.random.binomial(1, prob, size=(n, 1))
    sig = np.identity(p)
    gauss_noise = np.random.multivariate_normal(np.zeros(p), sig, size=n)

    x = np.random.binomial(1, prob, size=(n, p)) + 0.5 * noise * gauss_noise
    coeffs = _gen_coeffs(p)

    eps = _calc_eps(n)
    x_coeffs = x @ coeffs
    y = binom * 2 - 1
    y = np.multiply(x_coeffs, y) + 0.5 * noise * eps

    return x, y


def logarithmic(n, p, noise=False):
    r"""
    Logarithmic simulation.

    Logarithmic :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}^p`:
    :math:`\epsilon \sim \mathcal{N}(0, I_p)`,

    .. math::

        X &\sim \mathcal{N}(0, I_p) \\
        Y_{|d|} &= 2 \log_2 (|X_{|d|}|) + 3 \kappa \epsilon_{|d|}
                   \ \mathrm{for}\ d = 1, ..., p

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, p)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    extra_args = {
        "noise": (noise, bool),
    }
    check_in = _CheckInputs(n, p)
    check_in(**extra_args)

    sig = np.identity(p)
    x = np.random.multivariate_normal(np.zeros(p), sig, size=n)
    eps = _calc_eps(n)

    y = np.log(x**2) + 3 * noise * eps

    return x, y


def fourth_root(n, p, noise=False, low=-1, high=1):
    r"""
    Fourth Root simulation.

    Fourth Root :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}`:

    .. math::

        X &\sim \mathcal{U}(-1, 1)^p \\
        Y &= |w^T X|^\frac{1}{4} + \frac{\kappa}{4} \epsilon

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.
    low : float, default: -1
        The lower limit of the uniform distribution simulated from.
    high : float, default: 1
        The upper limit of the uniform distribution simulated from.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, 1)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    extra_args = {
        "noise": (noise, bool),
        "low": (low, float),
        "high": (high, float),
    }
    check_in = _CheckInputs(n, p)
    check_in(**extra_args)

    x = _random_uniform(n, p, low, high)
    eps = _calc_eps(n)
    coeffs = _gen_coeffs(p)

    x_coeffs = x @ coeffs
    y = np.abs(x_coeffs) ** 0.25 + 0.25 * noise * eps

    return x, y


def _sin(n, p, noise=False, low=-1, high=1, period=4 * np.pi):
    """Helper function to calculate sine simulation"""
    extra_args = {
        "noise": (noise, bool),
        "low": (low, float),
        "high": (high, float),
        "period": (period, float),
    }
    check_in = _CheckInputs(n, p)
    check_in(**extra_args)

    x = _random_uniform(n, p, low, high)
    if p > 1 or noise:
        sig = np.identity(p)
        v = np.random.multivariate_normal(np.zeros(p), sig, size=n)
        x = x + 0.02 * p * v
    eps = _calc_eps(n)

    if period == 4 * np.pi:
        cc = 1
    else:
        cc = 0.5

    y = np.sin(x * period) + cc * noise * eps

    return x, y


def sin_four_pi(n, p, noise=False, low=-1, high=1):
    r"""
    Sine 4\ :math:`\pi` simulation.

    Sine 4\ :math:`\pi` :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}^p`:
    :math:`U \sim \mathcal{U}(-1, 1)`, :math:`V \sim \mathcal{N}(0, 1)^p`,
    :math:`\theta = 4 \pi`,

    .. math::

        X_{|d|} &= U + 0.02 p V_{|d|}\ \mathrm{for}\ d = 1, ..., p \\
        Y &= \sin (\theta X) + \kappa \epsilon

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.
    low : float, default: -1
        The lower limit of the uniform distribution simulated from.
    high : float, default: 1
        The upper limit of the uniform distribution simulated from.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, p)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    return _sin(n, p, noise=noise, low=low, high=high, period=4 * np.pi)


def sin_sixteen_pi(n, p, noise=False, low=-1, high=1):
    r"""
    Sine 16\ :math:`\pi` simulation.

    Sine 16\ :math:`\pi` :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}^p`:
    :math:`U \sim \mathcal{U}(-1, 1)`, :math:`V \sim \mathcal{N}(0, 1)^p`,
    :math:`\theta = 16 \pi`,

    .. math::

        X_{|d|} &= U + 0.02 p V_{|d|}\ \mathrm{for}\ d = 1, ..., p \\
        Y &= \sin (\theta X) + \kappa \epsilon

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.
    low : float, default: 1
        The lower limit of the uniform distribution simulated from.
    high : float, default: 1
        The upper limit of the uniform distribution simulated from.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, p)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    return _sin(n, p, noise=noise, low=low, high=high, period=16 * np.pi)


def _square_diamond(n, p, noise=False, low=-1, high=1, period=-np.pi / 2):
    """Helper function to calculate square/diamond simulation"""
    extra_args = {
        "noise": (noise, bool),
        "low": (low, float),
        "high": (high, float),
        "period": (period, float),
    }
    check_in = _CheckInputs(n, p)
    check_in(**extra_args)

    u = _random_uniform(n, p, low, high)
    v = _random_uniform(n, p, low, high)
    sig = np.identity(p)
    gauss_noise = np.random.multivariate_normal(np.zeros(p), sig, size=n)

    x = u * np.cos(period) + v * np.sin(period) + 0.05 * p * gauss_noise * noise
    y = -u * np.sin(period) + v * np.cos(period)

    return x, y


def square(n, p, noise=False, low=-1, high=1):
    r"""
    Square simulation.

    Square :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}^p`:
    :math:`U \sim \mathcal{U}(-1, 1)`, :math:`V \sim \mathcal{N}(0, 1)^p`,
    :math:`\theta = -\frac{\pi}{8}`,

    .. math::

        X_{|d|} &= U \cos(\theta) + V \sin(\theta) + 0.05 p \epsilon_{|d|}
        \ \mathrm{for}\ d = 1, ..., p \\
        Y_{|d|} &= -U \sin(\theta) + V \cos(\theta)

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.
    low : float, default: -1
        The lower limit of the uniform distribution simulated from.
    high : float, default: 1
        The upper limit of the uniform distribution simulated from.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, p)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    return _square_diamond(n, p, noise=noise, low=low, high=high, period=-np.pi / 8)


def two_parabolas(n, p, noise=False, low=-1, high=1, prob=0.5):
    r"""
    Two Parabolas simulation.

    Two Parabolas :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}^p`:

    .. math::

        X &\sim \mathcal{U}(-1, 1)^p \\
        Y &= ((w^T X)^2 + 2 \kappa \epsilon) \times \left( U = \frac{1}{2} \right)

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.
    low : float, default: -1
        The lower limit of the uniform distribution simulated from.
    high : float, default: 1
        The upper limit of the uniform distribution simulated from.
    prob : float, default: 0.5
        The probability of the bernoulli distribution simulated from.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, 1)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    extra_args = {
        "noise": (noise, bool),
        "low": (low, float),
        "high": (high, float),
        "prob": (prob, float),
    }
    check_in = _CheckInputs(n, p)
    check_in(**extra_args)

    x = _random_uniform(n, p, low, high)
    coeffs = _gen_coeffs(p)
    u = np.random.binomial(1, prob, size=(n, 1))
    rand_noise = _random_uniform(n, p, low=0, high=1)

    x_coeffs = x @ coeffs
    y = (x_coeffs**2 + 2 * noise * rand_noise) * (u - 0.5)

    return x, y


def _circle_ellipse(n, p, noise=False, low=-1, high=1, radius=1):
    """Helper function to calculate circle/ellipse simulation"""
    extra_args = {
        "noise": (noise, bool),
        "low": (low, float),
        "high": (high, float),
        "radius": (radius, float),
    }
    check_in = _CheckInputs(n, p)
    check_in(**extra_args)

    if p > 1:
        noise = True
    x = _random_uniform(n, p, low, high)
    rx = radius * np.ones((n, p))
    unif = _random_uniform(n, p, low, high)
    sig = np.identity(p)
    gauss_noise = np.random.multivariate_normal(np.zeros(p), sig, size=n)

    ry = np.ones((n, p))
    x[:, 0] = np.cos(unif[:, 0] * np.pi)
    for i in range(p - 1):
        x[:, i + 1] = x[:, i] * np.cos(unif[:, i + 1] * np.pi)
        x[:, i] = x[:, i] * np.sin(unif[:, i + 1] * np.pi)

    x = rx * x + 0.4 * noise * rx * gauss_noise
    y = ry * np.sin(unif[:, 0].reshape(n, 1) * np.pi)

    return x, y


def circle(n, p, noise=False, low=-1, high=1):
    r"""
    Circle simulation.

    Circle :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}^p`:
    :math:`U \sim \mathcal{U}(-1, 1)^p`, :math:`\epsilon \sim \mathcal{N}(0, I_p)`,
    :math:`r = 1`,

    .. math::

        X_{|d|} &= r \left( \sin(\pi U_{|d+1|}) \prod_{j=1}^d \cos(\pi U_{|j|})
        + 0.4 \epsilon_{|d|} \right)\ \mathrm{for}\ d = 1, ..., p-1 \\
        X_{|p|} &= r \left( \prod_{j=1}^p \cos(\pi U_{|j|})
        + 0.4 \epsilon_{|p|} \right) \\
        Y_{|d|} &= \sin(\pi U_{|1|})

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.
    low : float, default: -1
        The lower limit of the uniform distribution simulated from.
    high : float, default: 1
        The upper limit of the uniform distribution simulated from.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, p)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    return _circle_ellipse(n, p, noise=noise, low=low, high=high, radius=1)


def ellipse(n, p, noise=False, low=-1, high=1):
    r"""
    Ellipse simulation.

    Ellipse :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}^p`:
    :math:`U \sim \mathcal{U}(-1, 1)^p`, :math:`\epsilon \sim \mathcal{N}(0, I_p)`,
    :math:`r = 5`,

    .. math::

        X_{|d|} &= r \left( \sin(\pi U_{|d+1|}) \prod_{j=1}^d \cos(\pi U_{|j|})
        + 0.4 \epsilon_{|d|} \right)\ \mathrm{for}\ d = 1, ..., p-1 \\
        X_{|p|} &= r \left( \prod_{j=1}^p \cos(\pi U_{|j|})
        + 0.4 \epsilon_{|p|} \right) \\
        Y_{|d|} &= \sin(\pi U_{|1|})

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.
    low : float, default: -1
        The lower limit of the uniform distribution simulated from.
    high : float, default: 1
        The upper limit of the uniform distribution simulated from.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, p)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    return _circle_ellipse(n, p, noise=noise, low=low, high=high, radius=5)


def diamond(n, p, noise=False, low=-1, high=1):
    r"""
    Diamond simulation.

    Diamond :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}^p`:
    :math:`U \sim \mathcal{U}(-1, 1)`, :math:`V \sim \mathcal{N}(0, 1)^p`,
    :math:`\theta = -\frac{\pi}{4}`,

    .. math::

        X_{|d|} &= U \cos(\theta) + V \sin(\theta)
        + 0.05 p \epsilon_{|d|}\ \mathrm{for}\ d = 1, ..., p \\
        Y_{|d|} &= -U \sin(\theta) + V \cos(\theta)

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    noise : bool, default: False
        Whether or not to include noise in the simulation.
    low : float, default: -1
        The lower limit of the uniform distribution simulated from.
    high : float, default: 1
        The upper limit of the uniform distribution simulated from.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, p)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    return _square_diamond(n, p, noise=noise, low=low, high=high, period=-np.pi / 4)


def multiplicative_noise(n, p):
    r"""
    Multiplicative Noise simulation.

    Multiplicative Noise :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}^p`:
    :math:`\U \sim \mathcal{N}(0, I_p)`,

    .. math::

        X &\sim \mathcal{N}(0, I_p) \\
        Y_{|d|} &= U_{|d|} X_{|d|}\ \mathrm{for}\ d = 1, ..., p

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, p)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    check_in = _CheckInputs(n, p)
    check_in()

    sig = np.identity(p)
    x = np.random.multivariate_normal(np.zeros(p), sig, size=n)
    y = np.random.multivariate_normal(np.zeros(p), sig, size=n)
    y = np.multiply(x, y)

    return x, y


def multimodal_independence(n, p, prob=0.5, sep1=3, sep2=2):
    r"""
    Multimodal Independence data.

    Multimodal Independence :math:`(X, Y) \in \mathbb{R}^p \times \mathbb{R}^p`:
    :math:`U \sim \mathcal{N}(0, I_p)`, :math:`V \sim \mathcal{N}(0, I_p)`,
    :math:`U^\prime \sim \mathcal{B}(0.5)^p`, :math:`V^\prime \sim \mathcal{B}(0.5)^p`,

    .. math::

        X &= \frac{U}{3} + 2 U^\prime - 1 \\
        Y &= \frac{V}{3} + 2 V^\prime - 1

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    prob : float, default: 0.5
        The probability of the bernoulli distribution simulated from.
    sep1, sep2: float, default: 3, 2
        The separation between clusters of normally distributed data.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x` and ``y`` have shapes ``(n, p)`` and ``(n, p)``
        where `n` is the number of samples and `p` is the number of
        dimensions.
    """
    extra_args = {
        "prob": (prob, float),
        "sep1": (sep1, float),
        "sep2": (sep2, float),
    }
    check_in = _CheckInputs(n, p)
    check_in(**extra_args)

    sig = np.identity(p)
    u = np.random.multivariate_normal(np.zeros(p), sig, size=n)
    v = np.random.multivariate_normal(np.zeros(p), sig, size=n)
    u_2 = np.random.binomial(1, prob, size=(n, p))
    v_2 = np.random.binomial(1, prob, size=(n, p))

    x = u / sep1 + sep2 * u_2 - 1
    y = v / sep1 + sep2 * v_2 - 1

    return x, y


SIMULATIONS = {
    "linear": linear,
    "exponential": exponential,
    "cubic": cubic,
    "joint_normal": joint_normal,
    "step": step,
    "quadratic": quadratic,
    "w_shaped": w_shaped,
    "spiral": spiral,
    "uncorrelated_bernoulli": uncorrelated_bernoulli,
    "logarithmic": logarithmic,
    "fourth_root": fourth_root,
    "sin_four_pi": sin_four_pi,
    "sin_sixteen_pi": sin_sixteen_pi,
    "square": square,
    "two_parabolas": two_parabolas,
    "circle": circle,
    "ellipse": ellipse,
    "diamond": diamond,
    "multiplicative_noise": multiplicative_noise,
    "multimodal_independence": multimodal_independence,
}


def indep_sim(sim, n, p, **kwargs):
    r"""
    Independence simulation generator.

    Takes a simulation and the required parameters, and outputs the simulated
    data matrices.

    Parameters
    ----------
    sim : str
        The name of the simulation (from the :mod:`hyppo.tools` module) that is to be
        rotated.
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    **kwargs
        Additional keyword arguements for the desired simulation.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices.
    """
    if sim not in SIMULATIONS.keys():
        raise ValueError(
            "sim_name must be one of the following: {}".format(list(SIMULATIONS.keys()))
        )
    else:
        sim = SIMULATIONS[sim]

    return sim(n, p, **kwargs)



import numpy as np
import numpy.testing as testing

from ..fssd import (
    FSSD,
    ustat_h1_mean_variance,
    power_criterion,
    fssd_grid_search_kernel,
    FSSDH0SimCovObs,
    FSSDH0SimCovDraw,
)
from .._utils import meddistance, fit_gaussian_draw, constrain
from ..kernel import KGauss
from ..density import IsotropicNormal

from numpy.random import default_rng

import scipy.stats as stats
import pytest


class TestFSSD:
    @pytest.mark.parametrize("n", [100])
    @pytest.mark.parametrize("alpha", [0.01])
    @pytest.mark.parametrize("d", [1])
    @pytest.mark.parametrize("J", [1])
    def test_basic(self, n, alpha, d, J):
        seed = 12
        # sample
        mean = np.zeros(d)
        variance = 1
        isonorm = IsotropicNormal(mean, variance)

        # only one dimension of the mean is shifted
        draw_mean = mean + 0
        draw_variance = variance + 1

        rng = default_rng(seed)
        X = rng.standard_normal(size=(n, d)) * np.sqrt(draw_variance) + draw_mean

        # Test
        sig2 = meddistance(X, subsample=1000) ** 2
        k = KGauss(sig2)
        list_k = [k]

        # random test locations
        V = fit_gaussian_draw(X, J, seed=seed + 1)
        null_sim = FSSDH0SimCovObs(n_simulate=200, seed=3)
        extra_sim = FSSDH0SimCovDraw()
        fssd = FSSD(isonorm, k, V, null_sim=null_sim, alpha=alpha)
        check_sim = extra_sim.simulate(gof=fssd)
        power_criterion(p=isonorm, X=X, k=k, test_locs=V)
        fssd_grid_search_kernel(p=isonorm, X=X, test_locs=V, list_kernel=list_k)
        fssd.get_H1_mean_variance(X=X)

        tresult = fssd.test(X, return_simulated_stats=True)

        # assertions
        testing.assert_almost_equal(tresult["pvalue"], 0, decimal=1)
        testing.assert_almost_equal(tresult["test_stat"], 1.6, decimal=1)

    @pytest.mark.parametrize("n", [200])
    @pytest.mark.parametrize("alpha", [0.01])
    @pytest.mark.parametrize("d", [4])
    @pytest.mark.parametrize("J", [1])
    def test_ustat_h1_mean_variance(self, n, alpha, d, J):
        seed = 20
        constrain(20, 19, 21)
        # sample
        mean = np.zeros(d)
        variance = 1
        isonorm = IsotropicNormal(mean, variance)

        draw_mean = mean + 2
        draw_variance = variance + 1
        rng = default_rng(seed)
        X = rng.standard_normal(size=(n, d)) * np.sqrt(draw_variance) + draw_mean

        # Test
        sig_square = meddistance(X, subsample=1000) ** 2
        k = KGauss(sig_square)

        # random test locations
        V = fit_gaussian_draw(X, J, seed=seed + 1)

        null_sim = FSSDH0SimCovObs(n_simulate=200, seed=3)
        fssd = FSSD(isonorm, k, V, null_sim=null_sim, alpha=alpha)
        fea_tensor = fssd.feature_tensor(X)

        u_mean, u_variance = ustat_h1_mean_variance(fea_tensor)

        random_sim = FSSDH0SimCovObs()
        random_sim.simulate(gof=fssd, X=X)

        # assertions
        testing.assert_almost_equal(u_variance, 1, decimal=1)
        # should reject H0
        testing.assert_almost_equal(u_mean, 0.8, decimal=1)



"""
Module containing datasources for representing distributions.
"""
from __future__ import print_function, division

from builtins import range, object
from past.utils import old_div

from abc import ABC, abstractmethod
import autograd.numpy as np
import scipy.stats as stats
from numpy.random import default_rng


class DataSource(ABC):
    """
    A source of data allowing resampling. Subclasses may prefix
    class names with DS.
    """

    @abstractmethod
    def sample(self, n, seed):
        """Return a Data. Returned result should be deterministic given
        the input (n, seed)."""
        raise NotImplementedError()


class DSIsotropicNormal(DataSource):
    """
    A DataSource providing samples from a mulivariate isotropic normal
    distribution.
    """

    def __init__(self, mean, variance):
        """
        mean: a numpy array of length d for the mean
        variance: a positive floating-point number for the variance.
        """
        assert len(mean.shape) == 1
        self.mean = mean
        self.variance = variance

    def sample(self, n, seed=2):
        rng = default_rng(seed)
        d = len(self.mean)
        mean = self.mean
        variance = self.variance
        X = rng.standard_normal(size=(n, d)) * np.sqrt(variance) + mean
        return X


class DSNormal(DataSource):
    """
    A DataSource implementing a multivariate Gaussian.
    """

    def __init__(self, mean, cov):
        """
        mean: a numpy array of length d.
        cov: d x d numpy array for the covariance.
        """
        self.mean = mean
        self.cov = cov
        assert mean.shape[0] == cov.shape[0]
        assert cov.shape[0] == cov.shape[1]

    def sample(self, n, seed=3):
        rng = default_rng(seed)
        mvn = stats.multivariate_normal(self.mean, self.cov)
        X = mvn.rvs(size=n)
        if len(X.shape) == 1:
            # This can happen if d=1
            X = X[:, np.newaxis]
        return X



import numpy as np

from ..independence.dcorr import _dcov
from ..tools import compute_dist
from ._utils import _CheckInputs, k_sample_transform
from .base import KSampleTest, KSampleTestOutput
from .ksamp import KSample


class DISCO(KSampleTest):
    r"""
    Distance Components (DISCO) test statistic and p-value.

    DISCO is a powerful multivariate `k`-sample test. It leverages distance matrix
    capabilities (similar to tests like distance correlation or Dcorr). In fact, DISCO
    statistic is equivalent to our 2-sample formulation nonparametric MANOVA via
    independence testing, i.e. :class:`hyppo.ksample.KSample`,
    and to
    :class:`hyppo.independence.Dcorr`,
    :class:`hyppo.ksample.Energy`,
    :class:`hyppo.independence.Hsic`, and
    :class:`hyppo.ksample.MMD`
    :footcite:p:`pandaNonparMANOVAIndependence2021`
    :footcite:p:`shenExactEquivalenceDistance2020`.

    Parameters
    ----------
    compute_distance : str, callable, or None, default: "euclidean"
        A function that computes the distance among the samples within each
        data matrix.
        Valid strings for ``compute_distance`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,

            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.

        Set to ``None`` or ``"precomputed"`` if ``x`` and ``y`` are already distance
        matrices. To call a custom function, either create the distance matrix
        before-hand or create a function of the form ``metric(x, **kwargs)``
        where ``x`` is the data matrix for which pairwise distances are
        calculated and ``**kwargs`` are extra arguements to send to your custom
        function.
    bias : bool, default: False
        Whether or not to use the biased or unbiased test statistics.
    **kwargs
        Arbitrary keyword arguments for ``compute_distance``.

    Notes
    -----
    Traditionally, the formulation for the DISCO statistic
    is as follows :footcite:p:`szekelyTestingEqualDistributions`:

    Define
    :math:`\{ u^i_1 \stackrel{iid}{\sim} F_{U_1},\ i = 1, ..., n_1 \}` up to
    :math:`\{ u^j_k \stackrel{iid}{\sim} F_{V_1},\ j = 1, ..., n_k \}` as `k` groups
    of samples deriving from different distributions with the same
    dimensionality. If :math:`d(\cdot, \cdot)` is a distance metric (i.e. Euclidean),
    :math:`N = \sum_{i = 1}^k n_k`,
    and :math:`\mathrm{Energy}` is the Energy test statistic from
    :class:`hyppo.ksample.Energy`
    then,

    .. math::

        \mathrm{DISCO}_N(\mathbf{u}_1, \ldots, \mathbf{u}_k) =
        \sum_{1 \leq k < l \leq K} \frac{n_k n_l}{2N}
        \mathrm{Energy}_{n_k + n_l} (\mathbf{u}_k, \mathbf{u}_l)

    The implementation in the :class:`hyppo.ksample.KSample` class (using
    :class:`hyppo.independence.Dcorr`) is in
    fact equivalent to this implementation (for p-values) and statistics are
    equivalent up to a scaling factor :footcite:p:`pandaNonparMANOVAIndependence2021`.

    The p-value returned is calculated using a permutation test uses
    :meth:`hyppo.tools.perm_test`.
    The fast version of the test uses :meth:`hyppo.tools.chi2_approx`.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, compute_distance="euclidean", bias=False, **kwargs):
        KSampleTest.__init__(
            self, compute_distance=compute_distance, bias=bias, **kwargs
        )

    def statistic(self, *args):
        r"""
        Calulates the DISCO test statistic.

        Parameters
        ----------
        *args : ndarray of float
            Variable length input data matrices. All inputs must have the same
            number of samples and dimensions. That is, the shapes must be `(n, p)`
            where `n` are the number of samples and `p` is
            the number of dimensions.

        Returns
        -------
        stat : float
            The computed DISCO statistic.
        """
        inputs = list(args)
        N = [i.shape[0] for i in inputs]

        if len(set(N)) > 1:
            raise ValueError(
                "Shape mismatch, inputs must have same sample size, "
                "currently {}".format(len(set(N)))
            )
        u, v = k_sample_transform(inputs)

        distu, distv = compute_dist(u, v, metric=self.compute_distance, **self.kwargs)

        # exact equivalence transformation Dcov and DISCO
        stat = _dcov(distu, distv, self.bias) * np.sum(N) * len(N) / 2
        self.stat = stat

        return stat

    def test(self, *args, reps=1000, workers=1, auto=True, random_state=None):
        r"""
        Calculates the DISCO test statistic and p-value.

        Parameters
        ----------
        *args : ndarray of float
            Variable length input data matrices. All inputs must have the same
            number of samples and dimensions. That is, the shapes must be `(n, p)`
            where `n` is the number of samples and `p` is
            the number of dimensions.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.
        auto : bool, default: True
            Automatically uses fast approximation when `n` and size of array
            is greater than 20. If ``True``, and sample size is greater than 20, then
            :class:`hyppo.tools.chi2_approx` will be run. Parameters ``reps`` and
            ``workers`` are
            irrelevant in this case. Otherwise, :class:`hyppo.tools.perm_test` will be
            run.

        Returns
        -------
        stat : float
            The computed DISCO statistic.
        pvalue : float
            The computed DISCO p-value.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.ksample import DISCO
        >>> x = np.arange(7)
        >>> y = x
        >>> stat, pvalue = DISCO().test(x, y)
        >>> '%.3f, %.1f' % (stat, pvalue)
        '-1.566, 1.0'
        """
        inputs = list(args)
        check_input = _CheckInputs(
            inputs=inputs,
            indep_test="dcorr",
        )
        inputs = check_input()
        N = [i.shape[0] for i in inputs]

        if len(set(N)) > 1:
            raise ValueError(
                "Shape mismatch, inputs must have same sample size, "
                "currently {}".format(len(set(N)))
            )

        # observed statistic
        stat = self.statistic(*inputs)

        # since stat transformation is invariant under permutation, k-sample Dcorr
        # pvalue is identical to DISCO
        _, pvalue = KSample(
            indep_test="Dcorr",
            compute_distkern=self.compute_distance,
            bias=self.bias,
            **self.kwargs
        ).test(
            *inputs, reps=reps, workers=workers, auto=auto, random_state=random_state
        )

        return KSampleTestOutput(stat, pvalue)



from ._utils import meddistance, fit_gaussian_draw
from .fssd import FSSD, FSSDH0SimCovObs
from .datasource import DataSource
from .density import Normal, IsotropicNormal
from .kernel import KGauss

__all__ = [s for s in dir()]  # add imported tests to all

GOF_TESTS = {
    "fssd": FSSD,
}



from typing import NamedTuple

import numpy as np
from scipy.stats.distributions import chi2
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.metrics import pairwise_distances

from ._utils import _CheckInputs, sim_matrix
from .base import IndependenceTest
from .dcorr import _dcorr

FOREST_TYPES = {
    "classifier": RandomForestClassifier,
    "regressor": RandomForestRegressor,
}


class KMERFTestOutput(NamedTuple):
    stat: float
    pvalue: float
    kmerf_dict: dict


class KMERF(IndependenceTest):
    r"""
    Kernel Mean Embedding Random Forest (KMERF) test statistic and p-value.

    The KMERF test statistic is a kernel method for calculating independence by using
    a random forest induced similarity matrix as an input, and has been shown to have
    especially high gains in finite sample testing power in high dimensional settings
    :footcite:p:`shenLearningInterpretableCharacteristic2020`.

    Parameters
    ----------
    forest : "regressor", "classifier", default: "regressor"
        Type of forest used when running the independence test. If the `y` input in
        ``test`` is categorial, use the "classifier" keyword.
    ntrees : int, default: 500
        The number of trees used in the random forest.
    compute_distance : str, callable, or None, default: "euclidean"
        A function that computes the distance among the samples for `y`.
        Valid strings for ``compute_distance`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,

            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.

        Set to ``None`` or ``"precomputed"`` if ``y`` is already a distance
        matrices. To call a custom function, either create the distance matrix
        before-hand or create a function of the form ``metric(x, **kwargs)``
        where ``x`` is the data matrix for which pairwise distances are
        calculated and ``**kwargs`` are extra arguements to send to your custom
        function.
    distance_kwargs : dict
        Arbitrary keyword arguments for ``compute_distance``.
    **kwargs
        Additional arguments used for the forest (see
        :class:`sklearn.ensemble.RandomForestClassifier` or
        :class:`sklearn.ensemble.RandomForestRegressor`)

    Notes
    -----
    .. note::
        This algorithm is currently under review at a peer-reviewed conference.

    A description of KMERF in greater detail can be found in
    :footcite:p:`shenLearningInterpretableCharacteristic2020`. It is computed
    using the following steps:

    Let :math:`x` and :math:`y` be :math:`(n, p)` and :math:`(n, 1)` samples of random
    variables
    :math:`X` and :math:`Y`.

    + Run random forest with :math:`m` trees. Independent bootstrap samples of size
      :math:`n_{b} \leq n` are drawn to build a tree each time; each tree structure
      within the forest is denoted as :math:`\phi_w \in \mathbf{P}`,
      :math:`w \in \{ 1, \ldots, m \}`; :math:`\phi_w(x_i)` denotes the partition
      assigned to :math:`x_i`.

    + Calculate the proximity kernel:

      .. math::

         \mathbf{K}^{\mathbf{x}}_{ij} = \frac{1}{m} \sum_{w = 1}^{m} I(\phi_w(x_i)
         = \phi_w(x_j))

      where :math:`I(\cdot)`$` is the indicator function for how often two observations
      lie in the same partition.

    + Compute the induced kernel correlation: Let

      .. math::

         \mathbf{L}^{\mathbf{x}}_{ij}=
         \begin{cases}
             \mathbf{K}^{\mathbf{x}}_{ij}
             - \frac{1}{n-2} \sum_{t=1}^{n} \mathbf{K}^{\mathbf{x}}_{it}
             - \frac{1}{n-2} \sum_{s=1}^{n} \mathbf{K}^{\mathbf{x}}_{sj}
             + \frac{1}{(n-1)(n-2)} \sum_{s,t=1}^{n} \mathbf{K}^{\mathbf{x}}_{st}
             & \mbox{when} \ i \neq j \\
             0 & \mbox{ otherwise}
         \end{cases}

    + Then let :math:`\mathbf{K}^{\mathbf{y}}` be the Euclidean distance induced kernel,
      and similarly compute :math:`\mathbf{L}^{\mathbf{y}}` from
      :math:`\mathbf{K}^{\mathbf{y}}`. The unbiased kernel correlation equals

      .. math::

         \mathrm{KMERF}_n(\mathbf{x}, \mathbf{y}) = \frac{1}{n(n-3)}
         \mathrm{tr} \left( \mathbf{L}^{\mathbf{x}} \mathbf{L}^{\mathbf{y}} \right)

    The p-value returned is calculated using a permutation test using
    :meth:`hyppo.tools.perm_test`.

    References
    ----------
    .. footbibliography::
    """
    def __init__(self, forest="regressor", ntrees=500, **kwargs):
        self.is_ksamp = False
        if "is_ksamp" in kwargs.keys():
            del kwargs["is_ksamp"]
        if forest in FOREST_TYPES.keys():
            self.clf = FOREST_TYPES[forest](n_estimators=ntrees, **kwargs)
        else:
            raise ValueError("forest must be one of the following ")
        IndependenceTest.__init__(self)

    def statistic(self, x, y):
        r"""
        Helper function that calculates the KMERF test statistic.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, 1)`` where
            `n` is the number of samples and `p` is the number of
            dimensions.

        Returns
        -------
        stat : float
            The computed KMERF statistic.
        """
        self.clf.fit(x, y)
        self.distx = 1 - sim_matrix(self.clf, x)
        self.disty = pairwise_distances(y, metric="euclidean")
        self.stat = _dcorr(self.distx, self.disty, bias=False, is_fast=False)

        # get feature importances from gini-based random forest
        self.importances = self.clf.feature_importances_

        return self.stat

    def test(self, x, y, reps=1000, workers=1, auto=True, random_state=None):
        r"""
        Calculates the KMERF test statistic and p-value.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, 1)`` where
            `n` is the number of samples and `p` is the number of
            dimensions.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.
        auto : bool, default: True
            Automatically uses fast approximation when `n` and size of array
            is greater than 20. If ``True``, and sample size is greater than 20, then
            :class:`hyppo.tools.chi2_approx` will be run. Parameters ``reps`` and
            ``workers`` are
            irrelevant in this case. Otherwise, :class:`hyppo.tools.perm_test` will be
            run.

        Returns
        -------
        stat : float
            The computed KMERF statistic.
        pvalue : float
            The computed KMERF p-value.
        kmerf_dict : dict
            Contains additional useful returns containing the following keys:

                - feat_importance : ndarray of float
                    An array containing the importance of each dimension

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.independence import KMERF
        >>> x = np.arange(100)
        >>> y = x
        >>> '%.1f, %.2f' % KMERF().test(x, y)[:1] # doctest: +SKIP
        '1.0, 0.001'
        """
        check_input = _CheckInputs(x, y, reps=reps)
        x, y = check_input()

        if auto and x.shape[0] > 20:
            n = x.shape[0]
            stat = self.statistic(x, y)
            statx = _dcorr(self.distx, self.distx, bias=False, is_fast=False)
            staty = _dcorr(self.disty, self.disty, bias=False, is_fast=False)
            pvalue = chi2.sf(stat / np.sqrt(statx * staty) * n + 1, 1)
            self.stat = stat
            self.pvalue = pvalue
            self.null_dist = None
        else:
            stat, pvalue = super(KMERF, self).test(
                x, y, reps, workers, is_distsim=False, random_state=random_state
            )
        kmerf_dict = {"feat_importance": self.importances}

        return KMERFTestOutput(stat, pvalue, kmerf_dict)



import numpy as np
import pytest
from numpy.testing import assert_equal, assert_raises

from .. import SIMULATIONS, indep_sim


class TestIndepSimShape:
    @pytest.mark.parametrize("n", [100, 1000])
    @pytest.mark.parametrize("p", [1, 5])
    @pytest.mark.parametrize(
        "sim",
        SIMULATIONS.keys(),
    )
    def test_shapes(self, n, p, sim):
        np.random.seed(123456789)
        x, y = SIMULATIONS[sim](n, p)
        x1, y1 = indep_sim(sim, n, p)
        nx, px = x.shape
        ny, py = y.shape
        nx1, px1 = x1.shape
        ny1, py1 = y1.shape

        if sim in [
            "joint_normal",
            "logarithmic",
            "sin_four_pi",
            "sin_sixteen_pi",
            "two_parabolas",
            "square",
            "diamond",
            "circle",
            "ellipse",
            "multiplicative_noise",
            "multimodal_independence",
        ]:
            assert_equal(px, py)
            assert_equal(px1, py1)
        else:
            assert_equal(px, py * p)
            assert_equal(px1, py1 * p)
        assert_equal(nx, ny)
        assert_equal(nx1, ny1)


class TestIndepSimErrorWarn:
    """Tests errors and warnings."""

    def test_np_inctype(self):
        assert_raises(ValueError, indep_sim, n="a", p=1, sim="linear")
        assert_raises(ValueError, indep_sim, n=10, p=7.0, sim="linear")

    def test_low_n(self):
        assert_raises(ValueError, indep_sim, n=3, p=1, sim="linear")

    def test_low_p(self):
        assert_raises(ValueError, indep_sim, n=5, p=0, sim="linear")

    def test_extra_inctype(self):
        assert_raises(ValueError, indep_sim, n=5, p=0, sim="linear", low="a")

    def test_wrong_sim(self):
        assert_raises(ValueError, indep_sim, n=100, p=1, sim="abcd")

    def test_joint_lowcov(self):
        assert_raises(ValueError, indep_sim, n=100, p=20, sim="joint_normal")



from abc import ABC, abstractmethod
from typing import NamedTuple

from ..tools import perm_test


class IndependenceTestOutput(NamedTuple):
    stat: float
    pvalue: float


class IndependenceTest(ABC):
    r"""
    A base class for an independence test.

    Parameters
    ----------
    compute_distance : str, callable, or None, default: "euclidean" or "gaussian"
        A function that computes the distance among the samples within each
        data matrix.
        Valid strings for ``compute_distance`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,

            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.

        Alternatively, this function computes the kernel similarity among the
        samples within each data matrix.
        Valid strings for ``compute_kernel`` are, as defined in
        :func:`sklearn.metrics.pairwise.pairwise_kernels`,

            [``"additive_chi2"``, ``"chi2"``, ``"linear"``, ``"poly"``,
            ``"polynomial"``, ``"rbf"``,
            ``"laplacian"``, ``"sigmoid"``, ``"cosine"``]

        Note ``"rbf"`` and ``"gaussian"`` are the same metric.
    **kwargs
        Arbitrary keyword arguments for ``compute_distkern``.
    """

    def __init__(self, compute_distance=None, **kwargs):
        # set statistic and p-value
        self.stat = None
        self.pvalue = None
        self.compute_distance = compute_distance
        self.kwargs = kwargs

        super().__init__()

    @abstractmethod
    def statistic(self, x, y):
        r"""
        Calculates the independence test statistic.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions. Alternatively, ``x`` and ``y`` can be distance matrices,
            where the shapes must both be ``(n, n)``.
        """

    @abstractmethod
    def test(
        self,
        x,
        y,
        reps=1000,
        workers=1,
        is_distsim=True,
        perm_blocks=None,
        random_state=None,
    ):
        r"""
        Calculates the independence test statistic and p-value.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions. Alternatively, ``x`` and ``y`` can be distance matrices,
            where the shapes must both be ``(n, n)``.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.
        auto : bool, default: True
            Automatically uses fast approximation when `n` and size of array
            is greater than 20. If ``True``, and sample size is greater than 20, then
            :class:`hyppo.tools.chi2_approx` will be run. Parameters ``reps`` and
            ``workers`` are
            irrelevant in this case. Otherwise, :class:`hyppo.tools.perm_test` will be
            run.
        is_distsim : bool, default: True
            Whether or not ``x`` and ``y`` are input matrices.
        perm_blocks : None or ndarray, default: None
            Defines blocks of exchangeable samples during the permutation test.
            If None, all samples can be permuted with one another. Requires `n`
            rows. At each column, samples with matching column value are
            recursively partitioned into blocks of samples. Within each final
            block, samples are exchangeable. Blocks of samples from the same
            partition are also exchangeable between one another. If a column
            value is negative, that block is fixed and cannot be exchanged.

        Returns
        -------
        stat : float
            The computed independence test statistic.
        pvalue : float
            The computed independence p-value.
        """
        self.x = x
        self.y = y

        stat, pvalue, null_dist = perm_test(
            self.statistic,
            x,
            y,
            reps=reps,
            workers=workers,
            is_distsim=is_distsim,
            perm_blocks=perm_blocks,
            random_state=random_state,
        )
        self.stat = stat
        self.pvalue = pvalue
        self.null_dist = null_dist

        return IndependenceTestOutput(stat, pvalue)



from typing import NamedTuple

from ..independence import Dcorr
from ._utils import _CheckInputs, compute_stat
from .base import TimeSeriesTest


class DcorrXTestOutput(NamedTuple):
    stat: float
    pvalue: float
    dcorrx_dict: dict


class DcorrX(TimeSeriesTest):
    r"""
    Cross Distance Correlation (DcorrX) test statistic and p-value.

    DcorrX is an independence test between two (paired) time series of
    not necessarily equal dimensions. The population parameter is 0 if and only if the
    time series are independent. It is based upon energy distance between distributions.

    .. _[1]: https://arxiv.org/abs/1908.06486

    Parameters
    ----------
    compute_distance : str, callable, or None, default: "euclidean"
        A function that computes the distance among the samples within each
        data matrix.
        Valid strings for ``compute_distance`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,

            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.

        Set to ``None`` or ``"precomputed"`` if ``x`` and ``y`` are already distance
        matrices. To call a custom function, either create the distance matrix
        before-hand or create a function of the form ``metric(x, **kwargs)``
        where ``x`` is the data matrix for which pairwise distances are
        calculated and ``**kwargs`` are extra arguements to send to your custom
        function.
    max_lag : int, default: 0
        The maximum number of lags in the past to check dependence between ``x`` and the
        shifted ``y``. If ``None``, then ``max_lag=np.ceil(np.log(n))``. Also the
        ``M`` hyperparmeter below.
    **kwargs
        Arbitrary keyword arguments for ``compute_distance``.

    Notes
    -----
    The statistic can be derived as follows
    :footcite:p:`mehtaIndependenceTestingMultivariate2020`:

    Let :math:`x` and :math:`y` be :math:`(n, p)` and :math:`(n, q)` series
    respectively, which each contain :math:`y` observations of the series
    :math:`(X_t)` and :math:`(Y_t)`. Similarly, let :math:`x[j:n]` be the
    :math:`(n-j, p)` last :math:`n-j` observations of :math:`x`. Let :math:`y[0:(n-j)]`
    be the :math:`(n-j, p)` first :math:`n-j` observations of :math:`y`. Let :math:`M`
    be the maximum lag hyperparameter. The cross distance correlation is,

    .. math::

        \mathrm{DcorrX}_n (x, y) =  \sum_{j=0}^M \frac{n-j}{n}
                                    Dcorr_n (x[j:n], y[0:(n-j)])

    The p-value returned is calculated using a permutation test.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, compute_distance="euclidean", max_lag=0, **kwargs):
        TimeSeriesTest.__init__(
            self, compute_distance=compute_distance, max_lag=max_lag, **kwargs
        )

    def statistic(self, x, y):
        r"""
        Helper function that calculates the DcorrX test statistic.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions. Alternatively, ``x`` and ``y`` can be distance matrices,
            where the shapes must both be ``(n, n)``.

        Returns
        -------
        stat : float
            The computed DcorrX statistic.
        opt_lag : int
            The computed optimal lag.
        """
        stat, opt_lag = compute_stat(
            x, y, Dcorr, self.compute_distance, self.max_lag, **self.kwargs
        )
        self.stat = stat
        self.opt_lag = opt_lag

        return stat, opt_lag

    def test(self, x, y, reps=1000, workers=1, random_state=None):
        r"""
        Calculates the DcorrX test statistic and p-value.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions. Alternatively, ``x`` and ``y`` can be distance matrices,
            where the shapes must both be ``(n, n)``.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.

        Returns
        -------
        stat : float
            The computed DcorrX statistic.
        pvalue : float
            The computed DcorrX p-value.
        dcorrx_dict : dict
            Contains additional useful returns containing the following keys:

                - opt_lag : int
                    The optimal lag that maximizes the strength of the relationship.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.time_series import DcorrX
        >>> np.random.seed(456)
        >>> x = np.arange(7)
        >>> y = x
        >>> stat, pvalue, dcorrx_dict = DcorrX().test(x, y, reps = 100)
        >>> '%.1f, %.2f, %d' % (stat, pvalue, dcorrx_dict['opt_lag'])
        '1.0, 0.05, 0'
        """
        check_input = _CheckInputs(
            x,
            y,
            max_lag=self.max_lag,
        )
        x, y, self.max_lag = check_input()

        stat, pvalue, stat_list = super(DcorrX, self).test(
            x, y, reps, workers, random_state
        )
        dcorrx_dict = {"opt_lag": stat_list[1]}
        return DcorrXTestOutput(stat, pvalue, dcorrx_dict)



from abc import ABC, abstractmethod
from typing import NamedTuple


class ConditionalIndependenceTestOutput(NamedTuple):
    stat: float
    pvalue: float


class ConditionalIndependenceTest(ABC):
    """
    A base class for a conditional independence test.

    """

    def __init__(self, **kwargs):
        self.stat = None
        self.pvalue = None
        self.kwargs = kwargs

        super().__init__()

    @abstractmethod
    def statistic(self, x, y, z):
        r"""
        Calculates the conditional independence test statistic.

        Parameters
        ----------
        x,y,z : ndarray of float
            Input data matrices.

        Returns
        -------
        stat : float
            The computed conditional independence test statistic.
        """

    @abstractmethod
    def test(self, x, y, z):
        r"""
        Calculates the conditional independence test statistic and p-value.

        Parameters
        ----------
        x,y,z : ndarray of float
            Input data matrices.

        Returns
        -------
        stat : float
            The computed conditional independence test statistic.
        pvalue : float
            The computed conditional independence test p-value.
        """



"""
Module containing implementations of some unnormalized probability density 
functions.
"""
from __future__ import division

from builtins import range
from past.utils import old_div

from abc import ABC, abstractmethod
import autograd
import autograd.numpy as np
from .datasource import DSNormal, DSIsotropicNormal
import scipy.stats as stats


class UnnormalizedDensity(ABC):
    """
    An abstract class of an unnormalized probability density function.  This is
    intended to be used to represent a model of the data for goodness-of-fit
    testing.
    """
    
    @abstractmethod
    def log_den(self, X):
        """
        Evaluate this log of the unnormalized density on the n points in X.
        X: n x d numpy array
        Return a one-dimensional numpy array of length n.
        """
        raise NotImplementedError()

    def log_normalized_den(self, X):
        """
        Evaluate the exact normalized log density. The difference to log_den()
        is that this method adds the normalizer. This method is not
        compulsory. Subclasses do not need to override.
        """
        raise NotImplementedError()

    def get_datasource(self):
        """
        Return a DataSource that allows sampling from this density.
        May return None if no DataSource is implemented.
        Implementation of this method is not enforced in the subclasses.
        """
        return None

    def grad_log(self, X):
        """
        Evaluate the gradients (with respect to the input) of the log density at
        each of the n points in X. This is the score function. Given an
        implementation of log_den(), this method will automatically work.
        Subclasses may override this if a more efficient implementation is
        available.
        X: n x d numpy array.
        Return an n x d numpy array of gradients.
        """
        g = autograd.elementwise_grad(self.log_den)
        G = g(X)
        return G

    @abstractmethod
    def dim(self):
        """
        Return the dimension of the input.
        """
        raise NotImplementedError()


class IsotropicNormal(UnnormalizedDensity):
    """
    Unnormalized density of an isotropic multivariate normal distribution.
    """

    def __init__(self, mean, variance):
        """
        mean: a numpy array of length d for the mean
        variance: a positive floating-point number for the variance.
        """
        self.mean = mean
        self.variance = variance

    def log_den(self, X):
        mean = self.mean
        variance = self.variance
        unden = old_div(-np.sum((X - mean) ** 2, 1), (2.0 * variance))
        return unden

    def log_normalized_den(self, X):
        d = self.dim()
        return stats.multivariate_normal.logpdf(
            X, mean=self.mean, cov=self.variance * np.eye(d)
        )

    def get_datasource(self):
        return DSIsotropicNormal(self.mean, self.variance)

    def dim(self):
        return len(self.mean)


class Normal(UnnormalizedDensity):
    """
    A multivariate normal distribution.
    """

    def __init__(self, mean, cov):
        """
        mean: a numpy array of length d.
        cov: d x d numpy array for the covariance.
        """
        self.mean = mean
        self.cov = cov
        assert mean.shape[0] == cov.shape[0]
        assert cov.shape[0] == cov.shape[1]
        E, V = np.linalg.eigh(cov)
        if np.any(np.abs(E) <= 1e-7):
            raise ValueError("covariance matrix is not full rank.")
        # The precision matrix
        self.prec = np.dot(np.dot(V, np.diag(old_div(1.0, E))), V.T)

    def log_den(self, X):
        mean = self.mean
        X0 = X - mean
        X0prec = np.dot(X0, self.prec)
        unden = old_div(-np.sum(X0prec * X0, 1), 2.0)
        return unden

    def get_datasource(self):
        return DSNormal(self.mean, self.cov)

    def dim(self):
        return len(self.mean)



from .base import DVariateTest, DVariateTestOutput
from ..tools import multi_compute_kern
from ._utils import _CheckInputs

import numpy as np


class dHsic(DVariateTest):
    r"""
    :math:`d`-variate Hilbert Schmidt Independence Criterion (dHsic) test
    statistic and p-value.

    dHsic is a non-parametric kernel-based independence test between an
    arbitrary number of variables. The dHsic statistic is 0 if the variables
    are jointly independent and positive if the variables are dependent
    :footcite:p:`grettonKernelJointIndependence2016`.
    The default choice is the Gaussian kernel, which uses the median distance
    as the bandwidth, which is a characteristic kernel that guarantees that
    dHsic is a consistent test
    :footcite:p:`grettonKernelJointIndependence2016`
    :footcite:p:`grettonKernelStatisticalTest2007`
    :footcite:p:`grettonConsistentNonparametricTests2010`.

    Parameters
    ----------
    compute_kernel : str, callable, or None, default: "gaussian"
        A function that computes the kernel similarity among the samples within each
        data matrix.
        Valid strings for ``compute_kernel`` are, as defined in
        :func:`sklearn.metrics.pairwise.pairwise_kernels`,

            [``"additive_chi2"``, ``"chi2"``, ``"linear"``, ``"poly"``,
            ``"polynomial"``, ``"rbf"``,
            ``"laplacian"``, ``"sigmoid"``, ``"cosine"``]

        Note ``"rbf"`` and ``"gaussian"`` are the same metric.
        Set to ``None`` or ``"precomputed"`` if ``args`` are already similarity
        matrices. To call a custom function, either create the similarity matrix
        before-hand or create a function of the form :func:`metric(x, **kwargs)`
        where ``x`` is the data matrix for which pairwise kernel similarity matrices are
        calculated and kwargs are extra arguments to send to your custom function.
    bias : bool, default: False
        Whether or not to use the biased or unbiased test statistics.
    **kwargs
        Arbitrary keyword arguments for ``multi_compute_kern``.

    Notes
    -----
    The statistic can be derived as follows
    :footcite:p:`grettonKernelJointIndependence2016`:

    dHsic builds on the two-variable Hilbert Schmidt Independence Criterion (Hsic),
    implemented in :class:`hyppo.independence.Hsic`, but allows for an arbitrary
    number of variables. For a given kernel, the joint distribution and the product
    of the marginals is mapped to the reproducing kernel Hilbert space and the
    squared distance between the embeddings is calculated. The dHsic statistic can
    be calculated by,

    .. math::

        \mathrm{dHsic} (\mathbb{P}^{(X^1, ..., X^d)}) = \Big\Vert \Pi(\mathbb{P}^{X^1}
        \otimes \cdot\cdot\cdot \otimes \mathbb{P}^{X^d}) - \Pi(\mathbb{P}^
        {(X^1, ..., X^d)}) \Big\Vert^2_{\mathscr{H}}

    Similar to Hsic, dHsic uses a gaussian median kernel by default, and the p-value
    is calculated using a permutation test using :meth:`hyppo.tools.multi_perm_test`.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, compute_kernel="gaussian", bias=True, **kwargs):
        self.compute_kernel = compute_kernel
        self.bias = bias

        DVariateTest.__init__(self, compute_kernel=self.compute_kernel, **kwargs)

    def statistic(self, *args):
        """
        Helper function that calculates the dHsic test statistic.

        Parameters
        ----------
        *args: ndarray of float
            Variable length input data matrices. All inputs must have the same
            number of samples. That is, the shapes must be ``(n, p)``, ``(n, q)``,
            etc., where `n` is the number of samples and `p` and `q` are the
            number of dimensions.

        Returns
        -------
        stat : float
            The computed dHsic statistic.
        """
        kerns = multi_compute_kern(*args, metric=self.compute_kernel, **self.kwargs)

        n = kerns[0].shape[0]
        term1 = np.ones((n, n))
        term2 = 1
        term3 = (2 / n) * np.ones((n,))
        for kern in kerns:
            term1 = np.multiply(term1, kern)
            term2 = (1 / n**2) * term2 * np.sum(kern)
            term3 = (1 / n) * np.multiply(term3, np.sum(kern, axis=1))

        stat = (1 / n**2) * np.sum(term1) + term2 - np.sum(term3)
        self.stat = stat

        return stat

    def test(self, *args, reps=1000, workers=1):
        """
        Calculates the dHsic test statistic and p-value.

        Parameters
        ----------
        *args: ndarray of float
            Variable length input data matrices. All inputs must have the same
            number of samples. That is, the shapes must be ``(n, p)``, ``(n, q)``,
            etc., where `n` is the number of samples and `p` and `q` are the
            number of dimensions.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.

        Returns
        -------
        stat : float
            The computed dHsic statistic.
        pvalue : float
            The computed dHsic p-value.
        """
        check_input = _CheckInputs(
            *args,
            reps=reps,
        )
        args = check_input()

        stat, pvalue = super(dHsic, self).test(*args, reps=reps, workers=workers)
        self.stat = stat
        self.pvalue = pvalue

        return DVariateTestOutput(stat, pvalue)



import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_approx_equal, assert_raises

from ...tools import linear, multimodal_independence, power, spiral
from .. import KMERF


class TestKMERFStat(object):
    """Test validity of KMERF test statistic"""

    # commented out p-value calculation because build stalled
    @pytest.mark.parametrize(
        "sim, obs_stat, obs_pvalue",
        [
            (linear, 0.26096478961164743, 2.210668062138582e-07),  # test linear simulation
            (
                spiral,
                0.038750928574255245,
                0.031366764487060314,
            ),  # test spiral simulation
            (
                multimodal_independence,
                0.0,
                0.31731050786291115,
            ),  # test independence simulation
        ],
    )
    def test_oned(self, sim, obs_stat, obs_pvalue):
        np.random.seed(12345678)

        # generate x and y
        x, y = sim(n=100, p=1)

        # test stat and pvalue
        stat1 = KMERF().statistic(x, y)
        stat2, pvalue, _ = KMERF().test(x, y)
        assert_approx_equal(stat1, obs_stat, significant=1)
        assert_approx_equal(stat2, obs_stat, significant=1)
        assert_approx_equal(pvalue, obs_pvalue, significant=1)


class TestKmerfErrorWarn:
    """Tests errors and warnings derived from MGC."""

    def test_no_indeptest(self):
        # raises error if not indep test
        assert_raises(ValueError, KMERF, "abcd")


# code too slow to run, check benchmarks in docs for verification
@pytest.mark.skip
class TestKmerfTypeIError:
    def test_oned(self):
        np.random.seed(123456789)
        est_power = power(
            "KMERF",
            sim_type="indep",
            sim="multimodal_independence",
            n=5,
            p=1,
            alpha=0.05,
        )

        assert_almost_equal(est_power, 0.05, decimal=2)



import numpy as np
import pytest
from numpy.testing import assert_equal, assert_raises

from .. import condi_indep_sim, COND_SIMULATIONS


"""

class TestTSSimShape:
    @pytest.mark.parametrize("n", [100, 1000])
    @pytest.mark.parametrize(
        "sim",
        ["indep_ar", "cross_corr_ar", "nonlinear_process", "extinct_gaussian_process"],
    )
    def test_shapes(self, n, sim):
        np.random.seed(123456789)
        x, y = ts_sim(sim, n)
        assert_equal(x.shape, y.shape)


class TestTSSimErrorWarn:
    ""Tests errors and warnings."

    def test_np_inctype(self):
        assert_raises(ValueError, ts_sim, "indep_ar", n="a")

    def test_low_n(self):
        assert_raises(ValueError, ts_sim, "indep_ar", n=1)

    def test_extra_inctype(self):
        assert_raises(ValueError, ts_sim, "indep_ar", n=10, phi="a")

    def test_wrong_sim(self):
        assert_raises(ValueError, ts_sim, "abcd", n=10, phi=1)

"""


class TestCondiIndepSimShape:
    @pytest.mark.parametrize("n", [100, 1000])
    @pytest.mark.parametrize("p", [1, 5])
    @pytest.mark.parametrize(
        "sim",
        COND_SIMULATIONS.keys(),
    )
    def test_shapes(self, n, p, sim):
        np.random.seed(123456789)
        x, y, z = condi_indep_sim(sim, n, p)
        nx, px = x.shape
        ny, py = y.shape
        nz, pz = z.shape

        n = np.array([nx, ny, nz])

        assert np.all(n == nx)


class TestCondiIndepSimErrorWarn:
    """Tests errors and warnings."""

    def test_np_inctype(self):
        assert_raises(ValueError, condi_indep_sim, n="a", p=1, sim="independent_normal")
        assert_raises(
            ValueError, condi_indep_sim, n=10, p=7.0, sim="independent_normal"
        )

    def test_low_n(self):
        assert_raises(ValueError, condi_indep_sim, n=3, p=1, sim="independent_normal")

    def test_low_p(self):
        assert_raises(ValueError, condi_indep_sim, n=5, p=0, sim="independent_normal")

    def test_wrong_sim(self):
        assert_raises(ValueError, condi_indep_sim, n=100, p=1, sim="abcd")



import numpy as np
from hyppo.ksample._utils import _CheckInputs
from hyppo.ksample.base import KSampleTest, KSampleTestOutput
from scipy.stats import f


class Hotelling(KSampleTest):
    r"""
    Hotelling :math:`T^2` test statistic and p-value.

    Hotelling :math:`T^2` is 2-sample multivariate analysis of variance (MANOVA)
    and generalization of Student's t-test in arbitary dimension
    :footcite:p:`hotellingGeneralizationStudentRatio1931`.

    Notes
    -----
    The test statistic is formulated as below
    :footcite:p:`pandaNonparMANOVAIndependence2021`:

    Consider input samples :math:`u_i \stackrel{iid}{\sim} F_U` for
    :math:`i \in \{ 1, \ldots, n \}` and :math:`v_i \stackrel{iid}{\sim} F_V` for
    :math:`i \in \{ 1, \ldots, m \}`. Let :math:`\bar{u}` refer to the columnwise means
    of :math:`u`; that is, :math:`\bar{u} = (1/n) \sum_{i=1}^{n} u_i` and let
    :math:`\bar{v}` be the same for :math:`v`. Calculate sample covariance matrices
    :math:`\hat{\Sigma}_{uv} = u^T v` and sample variance matrices
    :math:`\hat{\Sigma}_{uu} = u^T u` and :math:`\hat{\Sigma}_{vv} = v^T v`.
    Denote pooled covariance matrix :math:`\hat{\Sigma}` as

    .. math::

       \hat{\Sigma} = \frac{(n -  1) \hat{\Sigma}_{uu} + (m - 1) \hat{\Sigma}_{vv} }
       {n + m - 2}

    Then,

    .. math::

       \text{Hotelling}_{n, m} (u, v) = \frac{n m}{n + m}
       (\bar{u} - \bar{v})^T \hat{\Sigma}^{-1} (\bar{u} - \bar{v})

    Since it is a multivariate generalization of Student's t-tests, it suffers from
    some of the same assumptions as Student's t-tests. That is, the validity of MANOVA
    depends on the assumption that random variables are normally distributed within
    each group and each with the same covariance matrix. Distributions of input data
    are generally not known and cannot always be reasonably modeled as Gaussian
    :footcite:p:`micceriUnicornNormalCurve1989`
    :footcite:p:`stiglerRobustEstimatorsWork1977`
    and having the same covariance across groups is also generally not true of
    real data.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self):
        KSampleTest.__init__(self)

    def statistic(self, x, y):
        r"""
        Calulates the Hotelling :math:`T^2` test statistic.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            dimensions. That is, the shapes must be ``(n, p)`` and ``(m, p)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions.

        Returns
        -------
        stat : float
            The computed Hotelling :math:`T^2` statistic.
        """
        # ported from Hotteling packge in R
        nx, p = x.shape
        ny = y.shape[0]

        meanx = np.mean(x, axis=0)
        meany = np.mean(y, axis=0)

        covx = np.cov(x, rowvar=False)
        covy = np.cov(y, rowvar=False)

        covs = ((nx - 1) * covx + (ny - 1) * covy) / (nx + ny - 2)
        m = (nx + ny - p - 1) / (p * (nx + ny - 2))
        if p > 1:
            inv_covs = np.linalg.pinv(covs)
            stat = (
                m * (meanx - meany).T @ inv_covs @ (meanx - meany) * nx * ny / (nx + ny)
            )
        else:
            inv_covs = 1 / p
            stat = m * (meanx - meany) ** 2 * inv_covs * nx * ny / (nx + ny)

        self.stat = stat

        return stat

    def test(self, x, y):
        r"""
        Calculates the Hotelling :math:`T^2` test statistic and p-value.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            dimensions. That is, the shapes must be ``(n, p)`` and ``(m, p)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions.

        Returns
        -------
        stat : float
            The computed Hotelling :math:`T^2` statistic.
        pvalue : float
            The computed Hotelling :math:`T^2` p-value.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.ksample import Hotelling
        >>> x = np.arange(7)
        >>> y = x
        >>> stat, pvalue = Hotelling().test(x, y)
        >>> '%.3f, %.1f' % (stat, pvalue)
        '0.000, 1.0'
        """
        check_input = _CheckInputs(
            inputs=[x, y],
        )
        x, y = check_input()

        stat = self.statistic(x, y)
        nx, p = x.shape
        ny = y.shape[0]
        pvalue = f.sf(stat, p, nx + ny - p - 1)
        self.stat = stat
        self.pvalue = pvalue
        self.null_dist = None

        return KSampleTestOutput(stat, pvalue)
